# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-27 21:18
from __future__ import unicode_literals

from django.db import migrations


def populate_videos(apps, schema_editor):
    Match = apps.get_model('competition', 'Match')
    for match in Match.objects.all():
        match.videos = [
            video.url
            for video in match.matchvideo_set.all()
        ]
        match.save()


def noop(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0026_matchvideo'),
    ]

    operations = [
        migrations.RunPython(populate_videos, noop),
    ]



from __future__ import unicode_literals

import logging

from django import forms
from django.conf import settings
from django.forms import widgets
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from froala_editor.widgets import FroalaEditor

from touchtechnology.common.forms.tz import (
    DAY_CHOICES,
    HOUR_CHOICES,
    MINUTE_CHOICES,
    MONTH_CHOICES,
    TIMEZONE_CHOICES,
)

try:
    from importlib import metadata
except ImportError:  # Python 3.6
    import importlib_metadata as metadata

logger = logging.getLogger(__name__)


class MultiWidget(widgets.MultiWidget):
    def build_attrs(self, *args, **kwargs):
        """
        In the underlying render process for the widget, build_attrs will be
        used to determine the ``final_attrs`` to be associated to each widget.

        Unfortunately in Django 1.10 it seems that "required" is being added
        and this is producing different rendered output than that of earlier
        releases.

        This method is used to restore the pre-1.10 behaviour so our tests can
        continue passing across versions 1.8-1.10.
        """
        import django

        attrs = super(MultiWidget, self).build_attrs(*args, **kwargs)
        if django.VERSION >= (1, 10):
            attrs.pop("required", None)
        return attrs


class HTMLWidget(FroalaEditor):
    def trigger_froala(self, *args, **kwargs):
        """
        We don't want to do inline scripts, we'll handle the loading once from
        froala.js, so return an empty string to overload the base FroalaEditor
        implementation.

        Note: required for django-froala-editor<3 AFAICT
        """
        version = metadata.version("django-froala-editor")
        if int(version.split(".")[0]) < 3:
            return ""
        return super().trigger_froala(*args, **kwargs)

    class Media:
        css = {
            "all": (
                "https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.3.0/"
                "codemirror.min.css",
                "touchtechnology/common/css/froala_themes/dark.min.css",
                "touchtechnology/common/css/froala_themes/gray.min.css",
                "touchtechnology/common/css/froala_themes/red.min.css",
                "touchtechnology/common/css/froala_themes/royal.min.css",
            ),
        }
        js = (
            "https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.3.0/"
            "codemirror.min.js",
            "touchtechnology/common/js/froala.js",
        )


class GoogleMapsWidget(MultiWidget):
    def __init__(self, height, width, zoom, attrs=None):
        self.attrs = attrs or {}
        self.height = height
        self.width = width
        self.zoom = zoom
        widgets = (
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
        )
        super(GoogleMapsWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return value.split(",")
        return ("", "", "")

    class Media:
        css = {"all": ("touchtechnology/common/css/location_widget.css",)}
        js = (
            "//maps.googleapis.com/maps/api/js?v=3",
            "touchtechnology/common/js/location_widget.js",
        )


class BootstrapGoogleMapsWidget(GoogleMapsWidget):
    def render(self, *args, **kwargs):
        base = super(BootstrapGoogleMapsWidget, self).render(*args, **kwargs)
        return """
            <div class="form-group">
                <label class="control-label col-sm-3">%s</label>
                <div class="col-sm-7">%s</div>
            </div>
        """ % (
            _("Location"),
            base,
        )


class SelectDateWidget(MultiWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.Select(choices=DAY_CHOICES, attrs=attrs),
            forms.Select(choices=MONTH_CHOICES, attrs=attrs),
            forms.TextInput(attrs=attrs),
        )
        super(SelectDateWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return (value.day, value.month, value.year)
        return (None, None, None)

    def format_output(self, rendered_widgets):
        return " ".join(rendered_widgets)


class SelectDateHiddenWidget(SelectDateWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
        )
        super(SelectDateWidget, self).__init__(widgets, attrs)


class SelectTimeWidget(MultiWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.Select(choices=HOUR_CHOICES, attrs=attrs),
            forms.Select(choices=MINUTE_CHOICES, attrs=attrs),
        )
        super(SelectTimeWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value is not None:
            return (value.hour, value.minute)
        return (None, None)

    def format_output(self, rendered_widgets):
        return " ".join(rendered_widgets)


class SelectTimeHiddenWidget(SelectTimeWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
        )
        super(SelectTimeWidget, self).__init__(widgets, attrs)


class SelectDateTimeWidget(MultiWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.Select(choices=DAY_CHOICES, attrs=attrs),
            forms.Select(choices=MONTH_CHOICES, attrs=attrs),
            forms.TextInput(attrs=attrs),
            forms.Select(choices=HOUR_CHOICES, attrs=attrs),
            forms.Select(choices=MINUTE_CHOICES, attrs=attrs),
        )
        if settings.USE_TZ:
            widgets += (forms.Select(choices=TIMEZONE_CHOICES, attrs=attrs),)
        super(SelectDateTimeWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if settings.USE_TZ:
            tzinfo = timezone.get_current_timezone()
            tzname = timezone.get_current_timezone_name()
            logger.debug("USE_TZ is True, timezone is %s (%s)", tzinfo, tzname)

            if value:
                logger.debug("Value is set (%s)", value)
                if timezone.is_naive(value):
                    value = timezone.make_aware(value, timezone.get_default_timezone())
                    logger.debug("Naive value, set default timezone (%s)", value)
                value = value.astimezone(tzinfo)
                logger.debug(
                    "Ensure timezone is correctly set to %s (%s)", tzname, value
                )
                values = (
                    value.day,
                    value.month,
                    value.year,
                    value.hour,
                    value.minute,
                    tzname,
                )
            else:
                values = (None, None, None, None, None, None)

        elif value:
            values = (value.day, value.month, value.year, value.hour, value.minute)

        else:
            values = (None, None, None, None, None)

        return values

    def format_output(self, rendered_widgets):
        date_part = " ".join(rendered_widgets[:3])
        time_part = " ".join(rendered_widgets[3:])
        output = """<span class="date_part">%(date_widget)s</span>
        <span class="time_part%(timezone)s">%(time_widget)s</span>""" % {
            "date_widget": date_part,
            "time_widget": time_part,
            "timezone": " tz" if settings.USE_TZ else "",
        }
        return output


class SelectDateTimeHiddenWidget(SelectDateTimeWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
        )
        if settings.USE_TZ:
            widgets += (forms.HiddenInput(),)
        super(SelectDateTimeWidget, self).__init__(widgets, attrs)



import factory
from django.db import transaction
from django.test.utils import override_settings
from test_plus import TestCase
from touchtechnology.news.tests import factories


@override_settings(ROOT_URLCONF="example_app.urls")
class InvalidDateTest(TestCase):
    def test_archive_day(self):
        self.get("news:article", year="2013", month="feb", day="31")
        self.response_404()

    def test_article(self):
        self.get(
            "news:article",
            year="2013",
            month="feb",
            day="31",
            slug="tfms-new-generaltechnical-manager",
        )
        self.response_404()


@override_settings(ROOT_URLCONF="example_app.urls")
class FeedTest(TestCase):
    def test_atom(self):
        self.assertGoodView("news:feed-atom")

    def test_rss(self):
        self.assertGoodView("news:feed-rss")


@override_settings(ROOT_URLCONF="example_app.urls")
class SiteTest(TestCase):
    def assertGoodArticleView(self, article, **kwargs):
        self.assertGoodView(
            "news:article",
            year=article.published.year,
            month=article.published.strftime("%b").lower(),
            day=article.published.day,
            slug=article.slug,
            **kwargs
        )

    def test_article(self):
        article = factories.ArticleFactory.create(
            headline="This is a predictable headline!"
        )
        self.assertEqual(article.slug, "this-is-a-predictable-headline")
        self.assertGoodArticleView(article)

    def test_article_one_category(self):
        categories = factories.CategoryFactory.create_batch(1)
        article = factories.ArticleFactory.create()
        article.categories.set(categories)
        self.assertGoodArticleView(article)

    def test_article_many_categories(self):
        categories = factories.CategoryFactory.create_batch(3)
        article = factories.ArticleFactory.create()
        article.categories.set(categories)
        self.assertGoodArticleView(article)

    def test_multiple_articles_related_categories(self):
        with transaction.atomic():
            categories = factories.CategoryFactory.create_batch(3)
            articles = factories.ArticleFactory.create_batch(2)
            for article in articles:
                article.categories.set(categories)
        self.assertGoodArticleView(article)

    def test_article_translation(self):
        article = factories.ArticleFactory.create(headline="This is in English")
        translation = factories.TranslationFactory.create(
            locale="de", article_id=article.pk, headline="Das ist in Deutsch"
        )
        with self.subTest(msg="Article references Translation"):
            self.assertGoodArticleView(article)
            self.assertResponseContains(
                '<a class="de" href="{}">Deutsch (German)</a>'.format(
                    translation.get_absolute_url()
                )
            )
        with self.subTest(msg="Translation references Article"):
            self.assertGoodView(
                "news:translation",
                year=article.published.year,
                month=article.published.strftime("%b").lower(),
                day=article.published.day,
                slug=article.slug,
                locale="de",
            )
            self.assertResponseContains(
                '<a href="{}">This is in English</a>'.format(article.get_absolute_url())
            )



from datetime import timedelta

from dateutil.parser import parse
from dateutil.rrule import DAILY, WEEKLY
from django import template
from django.apps import apps
from django.db.models import Case, F, Q, Sum, When
from django.template.loader import get_template
from django.utils import timezone
from first import first

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

register = template.Library()


@register.filter
def get_competitions(node):
    return node.get_children().filter(content_type__model="competition")


@register.filter
def opponent(match, team):
    if not match:
        return None
    elif match.home_team == team:
        return match.get_away_team()
    elif match.away_team == team:
        return match.get_home_team()
    else:
        return None


@register.inclusion_tag(
    "tournamentcontrol/competition/next_date.html", takes_context=True
)
def next_date(context, season, offset=0, datestr=None):
    from tournamentcontrol.competition.models import Match, Season

    if isinstance(season, str):
        season = Season.objects.get(pk=season)

    # TODO make the offset a value stored on the competition
    if datestr is None:
        now = timezone.now() - timedelta(minutes=offset)
    else:
        now = parse(datestr) - timedelta(minutes=offset)

    # drop any signifigance less than the minute
    now = now.replace(second=0, microsecond=0)

    # start with just matches for this season (and by definition competition)
    matches = Match.objects.filter(
        stage__division__draft=False, stage__division__season=season
    ).select_related(
        "play_at",
        "home_team__club",
        "away_team__club",
        "home_team__division__season__competition",
        "away_team__division__season__competition",
        "stage__division__season__competition",
    )

    # restrict to matches starting after "now"
    later_today = Q(date__exact=now.date(), time__gte=now.time())
    tomorrow_onwards = Q(date__gt=now.date())
    matches = matches.filter(later_today | tomorrow_onwards)

    try:
        next_game_date = first(matches.dates("date", "day").order_by("date"))
    except Match.DoesNotExist:
        matches = matches.none()
        next_game_date = None
    else:
        matches = matches.filter(date=next_game_date).exclude(is_bye=True)

    next_round = None

    if season.mode == DAILY:
        # For tournaments we want to update the upcoming matches every timeslot
        try:
            next_round = matches.filter(
                home_team_score__isnull=True, away_team_score__isnull=True
            )[0]
        except IndexError:
            matches = matches.none()
        else:
            matches = matches.filter(date=next_round.date, time=next_round.time)

    elif season.mode == WEEKLY:
        # For a weekly competition we want to update the upcoming matches each
        # week and sorted by division.
        matches = matches.order_by("stage__division", *Match._meta.ordering)

    context.update(
        {
            "matches": matches,
            "next_game_date": next_game_date,
            "next_round": next_round,
            "competition": season.competition,
            "season": season,
        }
    )

    return context


@register.simple_tag(takes_context=True)
def ladder(context, stage):
    if stage.pools.count():
        ordering = stage.ladder_summary.model._meta.ordering
        summary = stage.ladder_summary.order_by("stage_group", *ordering)
        template_name = "tournamentcontrol/competition/ladder/pool.html"
    else:
        summary = stage.ladder_summary.all()
        template_name = "tournamentcontrol/competition/ladder/standard.html"

    context.update(
        {"summary": summary,}
    )

    tpl = get_template(template_name)
    return tpl.render(context)


@register.simple_tag
def score(match, team, template_name="tournamentcontrol/competition/_score.html"):
    """
    Output a score relative to the specified team. Also includes the 'status'
    of the result from the perspective of the team.

        ie. Win 6-4
            Lose 3-5
            Forfeit 0-6

    Use the third argument to specify an alternative template file name.
    """
    if team == match.home_team:
        team_score = match.home_team_score
        opponent_score = match.away_team_score
    else:
        team_score = match.away_team_score
        opponent_score = match.home_team_score

    if team_score is not None and opponent_score is not None:
        if team_score > opponent_score or team == match.forfeit_winner:
            result = "won"
        elif team_score < opponent_score or match.is_forfeit:
            result = "lost"
        elif team_score == opponent_score:
            result = "drew"
        else:
            raise ValueError(
                "An error has occured while determining the " "match result."
            )
    else:
        result = None

    context = {
        "result": result,
        "team_score": team_score,
        "opponent_score": opponent_score,
        "forfeit": match.is_forfeit,
    }

    t = template.loader.get_template(template_name)

    return t.render(context)


@register.filter
def teams_in_season(team_queryset, season):
    assert isinstance(season, apps.get_model("competition", "season"))
    return team_queryset.filter(division__season=season)


@register.filter
def teams_in_division(team_queryset, division):
    assert isinstance(division, apps.get_model("competition", "division"))
    return team_queryset.filter(division=division)


@register.filter
def players(team, pad_to=None):
    try:
        players = list(team.people.filter(is_player=True))
    except AttributeError:
        players = []
    if pad_to is not None:
        while len(players) < pad_to:
            players.append(None)
    return players


@register.filter
def pair(i1, i2):
    return list(zip_longest(i1, i2))


@register.inclusion_tag("tournamentcontrol/competition/templatetags/statistics.html")
def statistics(match):
    stats = match.statistics.filter(played__gt=0).select_related("player")

    home_stats = stats.filter(
        match__home_team__people__team=match.home_team,
        player__teamassociation__team=match.home_team,
    ).distinct()

    away_stats = stats.filter(
        match__away_team__people__team=match.away_team,
        player__teamassociation__team=match.away_team,
    ).distinct()

    context = {
        "match": match,
        "lines": list(zip_longest(home_stats, away_stats)),
        "home_stats": home_stats,
        "away_stats": away_stats,
    }

    return context


@register.inclusion_tag("tournamentcontrol/competition/templatetags/preview.html")
def preview(match):
    annotate = {
        "played": Sum(
            Case(
                When(
                    person__statistics__match__stage__division=match.stage.division,
                    then=F("person__statistics__played"),
                ),
                default=0,
            )
        ),
        "points": Sum(
            Case(
                When(
                    person__statistics__match__stage__division=match.stage.division,
                    then=F("person__statistics__points"),
                ),
                default=0,
            )
        ),
    }

    home_team = match.home_team.people.filter(is_player=True).annotate(**annotate)
    away_team = match.away_team.people.filter(is_player=True).annotate(**annotate)

    context = {
        "match": match,
        "lines": list(zip_longest(home_team, away_team)),
        "home_team": home_team,
        "away_team": away_team,
    }

    return context



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-01-17 20:19
from __future__ import unicode_literals

from django.db import migrations


def link_person_model(apps, schema_editor):
    Club = apps.get_model('competition', 'Club')
    Person = apps.get_model('competition', 'Person')

    # Update the Club.primary_uuid field, with no reverse relation we need to
    # edit it on the object directly.
    for club in Club.objects.exclude(primary__isnull=True):
        club.primary_uuid = club.primary
        club.save()

    # Update all the reverse relations per person.
    for person in Person.objects.all():
        person.clubassociation_set.update(person_uuid=person.uuid)
        person.teamassociation_set.update(person_uuid=person.uuid)
        person.seasonassociation_set.update(person_uuid=person.uuid)
        person.statistics.update(player_uuid=person.uuid)


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0020_competition__add__person_uuid'),
    ]

    operations = [
        migrations.RunPython(
            link_person_model, reverse_code=migrations.RunPython.noop),
    ]



from __future__ import unicode_literals

import json
import re
from datetime import date, datetime, time

import pytz
from django import forms
from django.conf import settings
from django.db.models import Min
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.translation import gettext_lazy as _
from mptt.forms import TreeNodeChoiceField
from namedentities import named_entities

from touchtechnology.common.forms.iter import TemplateChoiceIterator
from touchtechnology.common.forms.mixins import LabelFromInstanceMixin
from touchtechnology.common.forms.widgets import (
    BootstrapGoogleMapsWidget,
    HTMLWidget,
    SelectDateHiddenWidget,
    SelectDateTimeHiddenWidget,
    SelectDateTimeWidget,
    SelectDateWidget,
    SelectTimeHiddenWidget,
    SelectTimeWidget,
)


class EmailField(forms.EmailField):
    """
    Custom EmailField which will transform email address to lowercase unless
    advised otherwise as this SHOULD NOT be significant.
    """

    def __init__(self, lowercase=True, *args, **kwargs):
        super(EmailField, self).__init__(*args, **kwargs)
        self.lowercase = lowercase

    def clean(self, data):
        data = super(EmailField, self).clean(data)
        if self.lowercase:
            data = data.lower()
        return data


class ModelMultipleChoiceField(LabelFromInstanceMixin, forms.ModelMultipleChoiceField):
    def __init__(self, select_related=True, *args, **kwargs):
        super(ModelMultipleChoiceField, self).__init__(*args, **kwargs)
        if select_related and self.queryset is not None:
            self.queryset = self.queryset.select_related()


class HTMLField(forms.CharField):
    widget = HTMLWidget

    def widget_attrs(self, widget):
        options = getattr(settings, "FROALA_EDITOR_OPTIONS", {})
        attrs = {
            "data-options": json.dumps(options),
        }
        return attrs

    def clean(self, value):
        if isinstance(value, str):
            value = named_entities(value)
        return super(HTMLField, self).clean(value)


class GoogleMapsField(forms.MultiValueField):
    def __init__(self, max_length, width=300, height=200, zoom=8, *args, **kwargs):
        fields = (
            forms.CharField(max_length=25, required=False),
            forms.CharField(max_length=25, required=False),
            forms.IntegerField(required=False),
        )
        kwargs["widget"] = BootstrapGoogleMapsWidget(height, width, zoom)
        super(GoogleMapsField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return ",".join([smart_str(d) for d in data_list])

    def clean(self, data):
        if [d for d in data if d]:
            return ",".join(data)
        return super(GoogleMapsField, self).clean(data)


class SelectDateField(forms.MultiValueField):
    widget = SelectDateWidget
    hidden_widget = SelectDateHiddenWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
        )
        super(SelectDateField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return date(data_list[2], data_list[1], data_list[0])

    def clean(self, data):
        if data is None:
            data = ["", "", ""]

        if not [d for d in data if d]:
            if self.required:
                raise forms.ValidationError("This field is required.")
            return None

        try:
            day, month, year = (int(i) for i in data)
            d = date(year, month, day)
        except ValueError:
            raise forms.ValidationError("Please enter a valid date.")
        return d


class SelectTimeField(forms.MultiValueField):
    widget = SelectTimeWidget
    hidden_widget = SelectTimeHiddenWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
        )
        super(SelectTimeField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return time(data_list[0], data_list[1])

    def clean(self, data):
        if not data:
            if self.required:
                raise forms.ValidationError("This field is required.")
            data = ("", "")

        t = None

        try:
            try:
                try:
                    hour, minute = data
                except ValueError:
                    raise forms.ValidationError(repr(data))

                try:
                    hour = int(hour)
                    assert 0 <= hour <= 23
                except (AssertionError, ValueError):
                    raise forms.ValidationError("Hour must be in 0..23")

                try:
                    minute = int(minute)
                    assert 0 <= minute <= 59
                except (AssertionError, ValueError):
                    raise forms.ValidationError("Minute must be in 0..59")

                try:
                    t = time(hour, minute)
                except ValueError as e:
                    raise forms.ValidationError(smart_str(e).capitalize())

            except TypeError as e:
                raise forms.ValidationError("Please enter a valid time.")

        except forms.ValidationError:
            if self.required or any(data):
                raise

        return t


class SelectDateTimeField(forms.MultiValueField):
    widget = SelectDateTimeWidget
    hidden_widget = SelectDateTimeHiddenWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
        )
        if settings.USE_TZ:
            fields += (forms.CharField(required=True),)
        super(SelectDateTimeField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return date(
            data_list[2], data_list[1], data_list[0], data_list[3], data_list[4]
        )

    def clean(self, data):
        if data is None:
            data = ["", "", "", "", ""]

        if not [d for d in data if d]:
            if self.required:
                raise forms.ValidationError("This field is required.")
            return None

        try:
            day, month, year, hour, minute = (int(i) for i in data[:5])
            value = datetime(year, month, day, hour, minute)
            if settings.USE_TZ:
                try:
                    tzinfo = pytz.timezone(data[5])
                except IndexError:
                    raise forms.ValidationError("Please select a time zone.")
                except pytz.exceptions.UnknownTimeZoneError:
                    raise forms.ValidationError("Please select a valid time " "zone.")
                value = timezone.make_aware(value, tzinfo)
        except ValueError:
            raise forms.ValidationError("Please enter a valid date and time.")

        return value


class TemplatePathFormField(forms.ChoiceField):
    def __init__(
        self,
        template_base,
        template_folder,
        match=None,
        recursive=False,
        required=True,
        widget=None,
        label=None,
        initial=None,
        help_text=None,
        cache_choices=False,
        empty_label=_("Default"),
        # This will never be used, but it's now a CharField in the
        # database so we need to pull it out of the keyword
        # arguments.
        max_length=None,
        *args,
        **kwargs
    ):

        super(TemplatePathFormField, self).__init__(
            required=required,
            widget=widget,
            label=label,
            initial=initial,
            help_text=help_text,
            *args,
            **kwargs
        )

        self.template_base = template_base
        self.template_folder = template_folder
        self.match = match
        self.recursive = recursive
        self.empty_label = empty_label

        if self.match is not None:
            self.match_re = re.compile(self.match)

        if not self.template_base.endswith("/"):
            self.template_base += "/"

        self.widget.choices = self.choices

    def _get_choices(self):
        return TemplateChoiceIterator(self)

    choices = property(_get_choices, forms.ChoiceField._set_choices)


class ModelChoiceField(LabelFromInstanceMixin, forms.ModelChoiceField):
    def __init__(self, select_related=True, *args, **kwargs):
        super(ModelChoiceField, self).__init__(*args, **kwargs)
        if select_related and self.queryset is not None:
            self.queryset = self.queryset.select_related()


class SitemapNodeModelChoiceField(ModelChoiceField):
    def __init__(self, label_from_instance="title", *args, **kwargs):
        super(SitemapNodeModelChoiceField, self).__init__(
            label_from_instance=label_from_instance, *args, **kwargs
        )


def boolean_choice_field_coerce(value):
    return bool(int(value))


class MinTreeNodeChoiceField(TreeNodeChoiceField):
    @property
    def minimum_level(self):
        if not hasattr(self, "_minimum_level"):
            self._minimum_level = self.queryset.aggregate(minimum=Min("level")).get(
                "minimum"
            )
        return self._minimum_level

    def label_from_instance(self, obj):
        level = obj.level - self.minimum_level
        return "%s %s" % (self.level_indicator * level, smart_str(obj))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-27 21:39
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0027_matchvideo_data'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='matchvideo',
            name='match',
        ),
        migrations.DeleteModel(
            name='MatchVideo',
        ),
    ]



# Generated by Django 2.2.7 on 2019-11-22 13:40

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0035_season_disable_calendar"),
    ]

    operations = [
        migrations.AlterField(
            model_name="divisionexclusiondate",
            name="division",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="exclusions",
                to="competition.Division",
            ),
        ),
        migrations.AlterField(
            model_name="ground",
            name="place_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="competition.Place",
            ),
        ),
        migrations.AlterField(
            model_name="ladderentry",
            name="match",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ladder_entries",
                to="competition.Match",
            ),
        ),
        migrations.AlterField(
            model_name="ladderentry",
            name="team",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ladder_entries",
                to="competition.Team",
            ),
        ),
        migrations.AlterField(
            model_name="laddersummary",
            name="stage",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ladder_summary",
                to="competition.Stage",
            ),
        ),
        migrations.AlterField(
            model_name="laddersummary",
            name="stage_group",
            field=touchtechnology.common.db.models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ladder_summary",
                to="competition.StageGroup",
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="gender",
            field=models.CharField(
                choices=[("M", "Male"), ("F", "Female"), ("X", "Unspecified")],
                max_length=1,
            ),
        ),
        migrations.AlterField(
            model_name="rankpoints",
            name="team",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="competition.RankTeam"
            ),
        ),
        migrations.AlterField(
            model_name="rankteam",
            name="club",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="competition.Club"
            ),
        ),
        migrations.AlterField(
            model_name="rankteam",
            name="division",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="competition.RankDivision",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="hashtag",
            field=models.CharField(
                blank=True,
                help_text="Your official <em>hash tag</em> for social media promotions.",
                max_length=30,
                null=True,
                validators=[
                    django.core.validators.RegexValidator(
                        "^(?:#)(\\w+)$",
                        "Enter a valid value. Make sure you include the # symbol.",
                    )
                ],
                verbose_name="Hash Tag",
            ),
        ),
        migrations.AlterField(
            model_name="seasonexclusiondate",
            name="season",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="exclusions",
                to="competition.Season",
            ),
        ),
        migrations.AlterField(
            model_name="seasonmatchtime",
            name="season",
            field=touchtechnology.common.db.models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="timeslots",
                to="competition.Season",
            ),
        ),
        migrations.AlterField(
            model_name="venue",
            name="place_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="competition.Place",
            ),
        ),
    ]



from __future__ import unicode_literals

import base64
import functools
import operator
from datetime import timedelta
from operator import or_

import pytz
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.contrib import messages
from django.contrib.sitemaps import views as sitemaps_views
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Case, Count, F, Q, Sum, When
from django.http import Http404, HttpResponse, HttpResponseGone
from django.shortcuts import get_object_or_404
from django.urls import include, path, re_path
from django.utils import timezone
from django.utils.encoding import smart_bytes
from django.utils.module_loading import import_string
from django.utils.translation import gettext, gettext_lazy as _
from guardian.utils import get_40x_or_None
from icalendar import Calendar, Event

from touchtechnology.common.decorators import login_required_m
from touchtechnology.common.sites import Application
from touchtechnology.common.utils import get_perms_for_model
from tournamentcontrol.competition.dashboard import (
    matches_require_basic_results,
    matches_require_details_results,
)
from tournamentcontrol.competition.decorators import competition_by_slug_m
from tournamentcontrol.competition.forms import (
    ConfigurationForm,
    DivisionTournamentScheduleForm,
    MatchResultFormSet,
    MatchStatisticFormset,
    MultiConfigurationForm,
    RankingConfigurationForm,
)
from tournamentcontrol.competition.models import (
    Competition,
    Match,
    Person,
    SimpleScoreMatchStatistic,
)
from tournamentcontrol.competition.rank import (
    DayView as RankDay,
    DivisionView as RankDivision,
    IndexView as RankIndex,
    MonthView as RankMonth,
    TeamView as RankTeam,
    YearView as RankYear,
)
from tournamentcontrol.competition.utils import (
    FauxQueryset,
    team_needs_progressing,
)


def permissions_required(
    request,
    model,
    instance=None,
    return_403=True,
    accept_global_perms=True,
    create=False,
    perms=None,
):
    # If no perms are specified, build sensible default using built in
    # permission types that come batteries included with Django.
    if perms is None:
        perms = get_perms_for_model(model, change=True)

        # When we're doing a creation we should have permission to create the object.
        if create:
            perms = get_perms_for_model(model, add=True)

    # Determine the user's permission to edit this object using the
    # get_40x_or_None - saves decorating view method with
    # @permission_required_or_403
    has_permission = get_40x_or_None(
        request,
        perms,
        obj=instance,
        return_403=return_403,
        accept_global_perms=accept_global_perms,
    )

    if has_permission is not None:
        return has_permission


class CompetitionAdminMixin(object):
    """
    Some administrative functions should be allowed to authenticated users on
    the front-end of the site. As long as we only use primitives from the base
    ``touchtechnology.common.sites.Application`` class it will integrate fine
    with both.
    """

    def day_runsheet(
        self, request, season, date, extra_context, templates=None, **kwargs
    ):
        matches = season.matches.filter(date=date).order_by(
            "is_bye", "datetime", "play_at"
        )
        templates = self.template_path(
            "runsheet.html", season.slug, season.competition.slug
        )
        return self.generic_list(
            request,
            matches,
            templates=templates,
            paginate_by=0,
            permission_required=False,
            extra_context=extra_context,
        )

    def match_results(
        self,
        request,
        competition,
        season,
        date,
        extra_context,
        redirect_to,
        division=None,
        time=None,
        **kwargs
    ):
        matches = Match.objects.filter(
            stage__division__season=season,
            stage__division__season__competition=competition,
            date=date,
        )

        if division is not None:
            matches = matches.filter(stage__division=division)

        # ensure only matches with progressed teams are able to be updated
        matches = matches.exclude(team_needs_progressing, is_bye=False)
        matches = matches.exclude(home_team__isnull=True, away_team__isnull=True)

        # FIXME too complex, be verbose so we can all read and understand it
        if time is not None:
            base = Q(time=time)
            bye_kwargs = matches.filter(time=time).values("stage", "round")
            time_or_byes = functools.reduce(
                operator.or_, [Q(time__isnull=True, **kw) for kw in bye_kwargs], base
            )
            matches = matches.filter(time_or_byes)
        else:
            matches = matches.order_by("date", "time", "play_at")

        match_queryset = matches.filter(is_bye=False)
        bye_queryset = matches.filter(is_bye=True)

        if request.method == "POST":
            match_formset = MatchResultFormSet(
                data=request.POST, queryset=match_queryset, prefix="matches"
            )
            bye_formset = MatchResultFormSet(
                data=request.POST, queryset=bye_queryset, prefix="byes"
            )

            if match_formset.is_valid() and bye_formset.is_valid():
                match_formset.save()
                bye_formset.save()
                messages.success(request, _("Your changes have been saved."))
                return self.redirect(redirect_to)
        else:
            match_formset = MatchResultFormSet(
                queryset=match_queryset, prefix="matches"
            )
            bye_formset = MatchResultFormSet(queryset=bye_queryset, prefix="byes")

        context = {
            "competition": competition,
            "season": season,
            "date": date,
            "division": division,
            "match_formset": match_formset,
            "bye_formset": bye_formset,
            "matches": matches,
            "cancel_url": redirect_to,
        }
        context.update(extra_context)

        templates = self.template_path("match_results.html")
        return self.render(request, templates, context)

    def edit_match_detail(
        self, request, stage, match, extra_context, redirect_to, **kwargs
    ):
        conditions = {
            "home_team__isnull": False,
            "away_team__isnull": False,
            "home_team_score__isnull": False,
            "away_team_score__isnull": False,
        }

        if match is None:
            match = get_object_or_404(stage.matches, pk=match.pk, **conditions)

        def team_faux_queryset(team):
            stats = FauxQueryset(SimpleScoreMatchStatistic, team=team)
            for player in team.people.filter(is_player=True):
                try:
                    statistic = SimpleScoreMatchStatistic.objects.get(
                        match=match, player=player.person
                    )
                except ObjectDoesNotExist:
                    statistic = SimpleScoreMatchStatistic(
                        match=match,
                        player=player.person,
                        number=player.number,
                        played=1,
                    )
                stats.append(statistic)
            return stats

        home_queryset = team_faux_queryset(match.home_team)
        away_queryset = team_faux_queryset(match.away_team)

        if request.method == "POST":
            home = MatchStatisticFormset(
                data=request.POST,
                score=match.home_team_score,
                prefix="home",
                queryset=home_queryset,
            )

            away = MatchStatisticFormset(
                data=request.POST,
                score=match.away_team_score,
                prefix="away",
                queryset=away_queryset,
            )

            if home.is_valid() and away.is_valid():
                home.save()
                away.save()
                messages.success(request, _("Your changes have been saved."))
                return self.redirect(redirect_to)

        else:
            home = MatchStatisticFormset(
                prefix="home", score=match.home_team_score, queryset=home_queryset
            )
            away = MatchStatisticFormset(
                prefix="away", score=match.away_team_score, queryset=away_queryset
            )

        context = {
            "object": match,
            "formsets": (home, away),
        }
        context.update(extra_context)

        templates = self.template_path("match_detail.html")
        return self.render(request, templates, context)


class CompetitionSite(CompetitionAdminMixin, Application):
    kwargs_form_class = ConfigurationForm

    def __init__(self, name="competition", app_name="competition", **kwargs):
        # store the node for future reference
        self.node = kwargs.get("node")
        super().__init__(name=name, app_name=app_name, **kwargs)
        self._competitions = Competition.objects.filter(enabled=True)
        if "competition" in kwargs and kwargs["competition"]:
            self._competitions = self._competitions.filter(slug=kwargs["competition"])

    def result_urls(self):
        return [
            path("", self.results, name="results"),
            path("match/<int:match>/", self.edit_match_detail, name="match-details"),
            re_path(r"^(?P<datestr>\d{8})/$", self.results, name="results"),
            re_path(
                r"^(?P<datestr>\d{8})/(?P<timestr>\d{4})/$",
                self.match_results,
                name="match-results",
            ),
        ]

    def runsheet_urls(self):
        return [
            path("", self.runsheet, name="runsheet"),
            re_path(r"^(?P<datestr>\d{8})/$", self.day_runsheet, name="runsheet"),
        ]

    def season_urls(self):
        return [
            path("", self.season, name="season"),
            path("forfeit/", self.forfeit_list, name="forfeit-list"),
            path("forfeit/<int:match>/", self.forfeit, name="forfeit"),
            path("videos/", self.season_videos, name="season-videos"),
            path("club:<slug:club>/", self.club, name="club"),
            path("club:<slug:club>.ics", self.calendar, name="calendar"),
            path("results/", include(self.result_urls())),
            path("runsheet/", include(self.runsheet_urls())),
            path("<slug:division>.ics", self.calendar, name="calendar"),
            path("<slug:division>/", self.division, name="division"),
            path("<slug:division>:<slug:stage>/", self.stage, name="stage"),
            path("<slug:division>:<slug:stage>:<slug:pool>/", self.pool, name="pool"),
            path("<slug:division>/match:<int:match>/", self.match, name="match"),
            path(
                "<slug:division>/match:<int:match>/video/",
                self.match_video,
                name="match-video",
            ),
            path("<slug:division>/<slug:team>.ics", self.calendar, name="calendar"),
            path("<slug:division>/<slug:team>/", self.team, name="team"),
        ]

    def competition_urls(self):
        return [
            path("", self.competition, name="competition"),
            path("<slug:season>.ics", self.calendar, name="calendar"),
            path("<slug:season>/", include(self.season_urls())),
        ]

    def get_urls(self):
        if self.kwargs.get("season") and self.kwargs.get("competition"):
            urlpatterns = [
                path("", include(self.season_urls()), kwargs=self.kwargs),
            ]
        elif self.kwargs.get("competition"):
            urlpatterns = [
                path("", include(self.competition_urls()), kwargs=self.kwargs),
            ]
        else:
            urlpatterns = [
                path("", self.index, name="index"),
                path("<slug:competition>/", include(self.competition_urls())),
            ]
        return urlpatterns

    def sitemap_index(
        self,
        request,
        node=None,
        competition=None,
        season=None,
        division=None,
        *args,
        **kwargs
    ):
        # We need to "pop" the node keyword argument
        return sitemaps_views.index(request, *args, **kwargs)

    def sitemap_section(
        self,
        request,
        node=None,
        competition=None,
        season=None,
        division=None,
        *args,
        **kwargs
    ):
        # We need to "pop" the node keyword argument
        return sitemaps_views.sitemap(request, *args, **kwargs)

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.name

    @property
    def competitions(self):
        return self._competitions.prefetch_related("seasons")

    def index(self, request, **kwargs):
        return self.generic_list(
            request,
            self._competitions,
            templates=self.template_path("index.html"),
            paginate_by=self._competitions.count(),
            extra_context=kwargs,
        )

    @competition_by_slug_m
    def competition(self, request, competition, extra_context, **kwargs):
        templates = self.template_path("competition.html", competition.slug)
        return self.generic_detail(
            request,
            self._competitions,
            slug=competition.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def season(self, request, competition, season, extra_context, **kwargs):
        templates = self.template_path("season.html", competition.slug, season.slug)
        extra_context.update(
            season.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage__division__season=season,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage__division__season=season,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            competition.seasons,
            slug=season.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def runsheet(self, request, competition, season, extra_context, **kwargs):
        context = {"dates": season.matches.dates("date", "day")}
        context.update(extra_context)
        templates = self.template_path(
            "runsheet_list.html", competition.slug, season.slug
        )
        return self.render(request, templates, context)

    @competition_by_slug_m
    def day_runsheet(self, request, season, date, extra_context, **kwargs):
        return super().day_runsheet(request, season, date, extra_context, **kwargs)

    @competition_by_slug_m
    @login_required_m
    def results(self, request, competition, season, extra_context, date=None, **kwargs):
        has_permission = permissions_required(request, Match, return_403=False)
        if has_permission is not None:
            return has_permission

        if date is not None:
            matches = season.matches.filter(date=date)
        else:
            matches = season.matches

        # Allow super-user accounts visibility to look 1 year into the future.
        if request.user.is_superuser:
            now = timezone.now() + relativedelta(years=1)
        else:
            now = None

        # Filter the list of matches to those which require result entry
        basic_results = matches_require_basic_results(now, matches)
        details_results = matches_require_details_results(matches, True)

        context = {
            "datetimes": [
                timezone.localtime(each, pytz.timezone(season.timezone))
                for each in basic_results.datetimes("datetime", "minute")
            ],
            "details": details_results,
        }
        context.update(extra_context)
        templates = self.template_path("results.html", competition.slug, season.slug)
        return self.render(request, templates, context)

    @competition_by_slug_m
    @login_required_m
    def match_results(
        self, request, competition, season, date, time, extra_context, **kwargs
    ):
        has_permission = permissions_required(request, Match, return_403=False)
        if has_permission is not None:
            return has_permission
        redirect_to = self.reverse("results", args=(competition.slug, season.slug))
        return super().match_results(
            request,
            competition=competition,
            season=season,
            date=date,
            time=time,
            extra_context=extra_context,
            redirect_to=redirect_to,
            **kwargs
        )

    @login_required_m
    @competition_by_slug_m
    def edit_match_detail(self, request, match, extra_context, **kwargs):
        has_permission = permissions_required(
            request, Match, instance=match, return_403=False
        )
        if has_permission is not None:
            return has_permission
        redirect_to = self.reverse(
            "results",
            args=(
                match.stage.division.season.competition.slug,
                match.stage.division.season.slug,
            ),
        )
        return super().edit_match_detail(
            request,
            stage=match.stage,
            match=match,
            extra_context=extra_context,
            redirect_to=redirect_to,
            **kwargs
        )

    @competition_by_slug_m
    def season_videos(self, request, competition, season, extra_context, **kwargs):
        templates = self.template_path(
            "season_videos.html", competition.slug, season.slug
        )
        queryset = (
            Match.objects.select_related(
                "stage__division",
                "home_team__club",
                "away_team__club",
                "play_at",
            )
            .exclude(videos__isnull=True)
            .filter(stage__division__season=season)
            .order_by("datetime", "play_at")
        )
        return self.generic_list(
            request,
            queryset,
            paginate_by=1000,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def club(self, request, competition, season, club, extra_context, **kwargs):
        templates = self.template_path(
            "club.html", competition.slug, season.slug, club.slug
        )
        return self.generic_detail(
            request,
            season.clubs,
            slug=club.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def division(self, request, competition, season, division, extra_context, **kwargs):
        templates = self.template_path(
            "division.html", competition.slug, season.slug, division.slug
        )
        extra_context.update(
            division.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage__division=division,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage__division=division,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            season.divisions,
            slug=division.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def stage(
        self, request, competition, season, division, stage, extra_context, **kwargs
    ):
        templates = self.template_path(
            "stage.html", competition.slug, season.slug, division.slug, stage.slug
        )
        extra_context["parent"] = stage
        extra_context.update(
            stage.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage=stage, then=F("statistics__points")
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage=stage, then=F("statistics__played")
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            division.stages,
            slug=stage.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def pool(
        self,
        request,
        competition,
        season,
        division,
        stage,
        pool,
        extra_context,
        **kwargs
    ):
        # FIXME sadly the template name was not changed when we refactored
        #       pools to be subordinates of Stage.
        templates = self.template_path(
            "divisiongroup.html",
            competition.slug,
            season.slug,
            division.slug,
            stage.slug,
            pool.slug,
        )
        extra_context.update(
            pool.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__match__stage_group=pool,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__match__stage_group=pool,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            stage.pools,
            slug=pool.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def team(
        self, request, competition, season, division, team, extra_context, **kwargs
    ):
        templates = self.template_path(
            "team.html", competition.slug, season.slug, division.slug, team.slug
        )

        extra_context.update(
            team.matches.exclude(is_bye=True).aggregate(
                timeslot_count=Count("datetime", distinct=True),
                match_count=Count("pk", distinct=True),
                points_scored=Sum(
                    Case(
                        When(
                            statistics__player__teamassociation__team=team,
                            then=F("statistics__points"),
                        )
                    )
                ),
                caps_awarded=Sum(
                    Case(
                        When(
                            statistics__player__teamassociation__team=team,
                            then=F("statistics__played"),
                        )
                    )
                ),
            )
        )
        return self.generic_detail(
            request,
            division.teams,
            slug=team.slug,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def calendar(self, request, season, club=None, division=None, team=None, **kwargs):
        if season.disable_calendar:
            # The GONE response informs client that they should remove this resource
            # from their cache. When a calendar has been added to user's mobile device
            # they may never look at it again, but we continue to process the requests
            # which can have poor performance. Try to influence a cleanup of clients.
            return HttpResponseGone()

        if team is not None:
            matches = team.matches
        elif division is not None:
            matches = division.matches
        elif club is not None:
            matches = club.matches.filter(stage__division__season=season)
        else:
            matches = season.matches

        # Do not include matches which have not had the time scheduled
        matches = matches.exclude(datetime__isnull=True)

        # Perform select_related to reduce extra queries
        matches = matches.select_related("stage__division__season__competition")

        # Reduce the size of the data set to return from the database
        matches = matches.defer(
            "date",
            "time",
            "home_team__copy",
            "home_team__names_locked",
            "home_team__order",
            "home_team__short_title",
            "home_team__timeslots_after",
            "home_team__timeslots_before",
            "home_team_score",
            "away_team__copy",
            "away_team__names_locked",
            "away_team__order",
            "away_team__short_title",
            "away_team__timeslots_after",
            "away_team__timeslots_before",
            "away_team_score",
            "bye_processed",
            "evaluated",
            "external_identifier",
            "forfeit_winner",
            "include_in_ladder",
            "is_bye",
            "is_forfeit",
            "is_washout",
            "rank_importance",
            "videos",
            "stage__division__season__competition__copy",
            "stage__division__season__competition__enabled",
            "stage__division__season__competition__order",
            "stage__division__season__competition__rank_importance",
            "stage__division__season__competition__short_title",
            "stage__division__season__competition__short_title",
            "stage__division__season__competition__slug_locked",
            "stage__division__season__competition__slug_locked",
            "stage__division__season__competition__title",
            "stage__division__season__complete",
            "stage__division__season__copy",
            "stage__division__season__disable_calendar",
            "stage__division__season__enabled",
            "stage__division__season__hashtag",
            "stage__division__season__mode",
            "stage__division__season__mvp_results_public",
            "stage__division__season__order",
            "stage__division__season__rank_importance",
            "stage__division__season__short_title",
            "stage__division__season__slug_locked",
            "stage__division__season__start_date",
            "stage__division__season__statistics",
            "stage__division__season__timezone",
            "stage__division__bonus_points_formula",
            "stage__division__copy",
            "stage__division__draft",
            "stage__division__forfeit_against_score",
            "stage__division__forfeit_for_score",
            "stage__division__games_per_day",
            "stage__division__include_forfeits_in_played",
            "stage__division__order",
            "stage__division__points_formula",
            "stage__division__rank_division",
            "stage__division__rank_importance",
            "stage__division__short_title",
            "stage__division__slug_locked",
            "stage__division__sportingpulse_url",
            "stage__carry_ladder",
            "stage__copy",
            "stage__follows",
            "stage__keep_ladder",
            "stage__keep_mvp",
            "stage__order",
            "stage__rank_importance",
            "stage__scale_group_points",
            "stage__short_title",
            "stage__slug_locked",
        )

        # Remove any matches that are part of a draft division unless being viewed
        # by a superuser.
        if not request.user.is_superuser:
            matches = matches.exclude(stage__division__draft=True)

        # For development server turn back plain text to make debugging easier
        if settings.DEBUG:
            content_type = "text/plain"
        else:
            content_type = "text/calendar"

        response = HttpResponse(content_type=content_type)

        cal = Calendar()
        cal.add("prodid", "-//Tournament Control//%s//" % request.get_host())
        cal.add("version", "2.0")

        for match in matches.order_by("datetime", "play_at"):
            # Determine the resource path to the detailed match view
            path = self.reverse(
                "match",
                kwargs={
                    "match": match.pk,
                    "division": match.stage.division.slug,
                    "season": match.stage.division.season.slug,
                    "competition": match.stage.division.season.competition.slug,
                },
            )

            # Combine the resource path with our current request context
            url = request.build_absolute_uri(str(path))

            event = Event()
            event["uid"] = "%s@%s" % (
                base64.b64encode(smart_bytes(url)),
                request.get_host(),
            )
            event.add("summary", match.title)
            event.add(
                "location",
                "{0} ({1})".format(match.stage.division.title, match.stage.title),
            )
            event.add("description", url)
            event.add("dtstart", match.datetime)
            # FIXME match duration should not be hardcoded
            event.add("dtend", match.datetime + timedelta(minutes=45))
            # FIXME should be the last modified time of the match
            event.add("dtstamp", timezone.now())

            cal.add_component(event)

        response.write(cal.to_ical())
        return response

    @competition_by_slug_m
    def match(
        self, request, competition, season, division, match, extra_context, **kwargs
    ):
        if match.is_bye or match.home_team is None or match.away_team is None:
            # Temporary, realistically nobody should ever have this URL but we
            # will use GONE so that Google removes it from their index.
            return HttpResponseGone()
        templates = self.template_path(
            "match.html", competition.slug, season.slug, division.slug
        )
        return self.generic_detail(
            request,
            division.matches,
            pk=match.pk,
            templates=templates,
            extra_context=extra_context,
        )

    @competition_by_slug_m
    def match_video(
        self, request, competition, season, division, match, extra_context, **kwargs
    ):
        if match.is_bye or match.home_team is None or match.away_team is None:
            # Temporary, realistically nobody should ever have this URL but we
            # will use GONE so that Google removes it from their index.
            return HttpResponseGone()
        templates = self.template_path(
            "videos.html", competition.slug, season.slug, division.slug
        )
        return self.generic_detail(
            request,
            division.matches,
            pk=match.pk,
            templates=templates,
            extra_context=extra_context,
        )

        # TODO add related model to record videos
        # return self.generic_list(request, match.videos,
        #                          templates=templates,
        #                          extra_context=extra_context)

    @login_required_m
    @competition_by_slug_m
    def forfeit_list(self, request, competition, season, extra_context, **kwargs):
        """
        List the matches that the visitor is permitted to forfeit. Must only
        show matches for teams that they are a registered member of (either as
        a player or other role).
        """
        try:
            teams = request.user.person.teams
        except Person.DoesNotExist:
            raise Http404

        matches = season.matches.filter(Q(home_team=teams) | Q(away_team=teams))

        context = {
            "matches": matches,
        }
        context.update(extra_context)

        templates = self.template_path("forfeit_list.html")
        return self.render(request, templates, context)

    @login_required_m
    @competition_by_slug_m
    def forfeit(self, request, competition, season, extra_context, match, **kwargs):
        """
        View that will allow a team member to forfeit a match they are due to
        be playing in.
        """
        try:
            teams = request.user.person.teams
        except Person.DoesNotExist:
            raise Http404

        matches = season.matches.filter(Q(home_team=teams) | Q(away_team=teams))

        redirect = self.redirect(
            self.reverse(
                "forfeit-list",
                kwargs={
                    "competition": competition.slug,
                    "season": season.slug,
                },
            )
        )

        if match not in matches:
            messages.add_message(
                request,
                messages.ERROR,
                "Attempting to forfeit a match you are not involved in.",
            )
            return redirect

        # decide which team to forfeit
        if match.home_team in teams and match.away_team in teams:
            messages.add_message(
                request,
                messages.WARNING,
                _(
                    "You are a member of both teams for this match, please "
                    "contact the competition manager."
                ),
            )
            return redirect

        elif match.home_team in teams:
            team = match.home_team
        elif match.away_team in teams:
            team = match.away_team
        else:
            messages.add_message(
                request,
                messages.ERROR,
                _(
                    "You are not a member of either team, please contact the "
                    "competition manager."
                ),
            )
            return redirect

        if request.method == "POST":
            # use the forfeit API to advise who made lodged the forfeit
            UNABLE_TO_POST = _(
                "Unable to post forfeit, please contact the " "competition manager."
            )
            try:
                success = match.forfeit(team, request.user.person)
            except AssertionError:
                # logger.exception("Failed precondition checks")
                messages.add_message(request, messages.ERROR, UNABLE_TO_POST)

            if success:
                messages.add_message(
                    request,
                    messages.INFO,
                    _(
                        "The match has been forfeit, notifications will be sent "
                        "to affected parties shortly."
                    ),
                )
            else:
                messages.add_message(request, messages.ERROR, UNABLE_TO_POST)

            return redirect

        context = {
            "match": match,
            "team": team,
        }
        context.update(extra_context)

        templates = self.template_path("forfeit.html")
        return self.render(request, templates, context)


class MultiCompetitionSite(CompetitionSite):
    kwargs_form_class = MultiConfigurationForm

    @classmethod
    def verbose_name(cls):
        return gettext("Multiple Competitions")

    def __init__(self, name="competition", app_name="competition", **kwargs):
        self.node = kwargs.get("node")  # store the node for future reference
        super().__init__(name=name, app_name=app_name, **kwargs)
        self._competitions = Competition.objects.filter(enabled=True)
        if "competition" in kwargs:
            q = functools.reduce(or_, [Q(slug=slug) for slug in kwargs["competition"]])
            self._competitions = self._competitions.filter(q)

    def get_urls(self):
        urlpatterns = [
            path("", self.index, name="index"),
            path("<slug:competition>/", include(self.competition_urls())),
        ]
        return urlpatterns


class RegistrationSite(Application):
    def __init__(self, name="rego", app_name="rego", **kwargs):
        super().__init__(name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        return []


class TournamentCalculatorSite(Application):
    def __init__(self, name="calculator", app_name="calculator", **kwargs):
        self.form_class = import_string(
            kwargs.pop(
                "form_class",
                "tournamentcontrol.competition.forms.TournamentScheduleForm",
            )
        )
        super().__init__(name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        urlpatterns = [
            path("", self.index, kwargs={"form_class": self.form_class}, name="index"),
            path(
                "live/",
                self.index,
                kwargs={
                    "form_class": DivisionTournamentScheduleForm,
                    "template_name": "live.html",
                },
                name="division",
            ),
        ]
        return urlpatterns

    def index(self, request, form_class, template_name="index.html", **extra_context):
        context = {}

        if request.GET:
            form = form_class(data=request.GET)
        else:
            form = form_class()

        if form.is_valid():
            context.setdefault("data", form.get_context())

        context.setdefault("form", form)
        context.update(extra_context)

        templates = self.template_path(template_name)
        return self.render(request, templates, context)


class RankingSite(Application):
    kwargs_form_class = RankingConfigurationForm

    def __init__(self, name="ranking", app_name="ranking", **kwargs):
        super().__init__(name=name, app_name=app_name, **kwargs)

    def get_urls(self):
        urlpatterns = [
            path("", RankIndex.as_view(), name="index"),
            path("<int:year>/", RankYear.as_view(), name="year"),
            path("<int:year>/<str:month>/", RankMonth.as_view(), name="month"),
            path("<int:year>/<str:month>/<int:day>/", RankDay.as_view(), name="day"),
            path(
                "<int:year>/<str:month>/<int:day>/<slug:slug>/",
                RankDivision.as_view(),
                name="rank",
            ),
            path(
                "<int:year>/<str:month>/<int:day>/<slug:slug>/<slug:team>/",
                RankTeam.as_view(),
                name="team",
            ),
        ]
        return urlpatterns


competition = CompetitionSite()
registration = RegistrationSite()
calculator = TournamentCalculatorSite()
ranking = RankingSite()



import os.path
from urllib.parse import urlunparse

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.urls import path, reverse
from django.utils.encoding import smart_str
from modelforms.forms import ModelForm

from touchtechnology.common.decorators import (
    never_cache_m,
    staff_login_required_m,
)
from touchtechnology.common.default_settings import SITEMAP_HTTPS_OPTION
from touchtechnology.common.forms.mixins import BootstrapFormControlMixin
from touchtechnology.common.sites import Application


class AdminComponentMixin(object):
    verbose_name = "ExampleComponent"
    visible = True

    # While adding this new functionality, we'll start out with
    # AdminComponents being unprotected by default, even though we typically
    # protect by @staff_login_required_m it's not our intent, we're moving
    # towards the permission's and django-guardian.
    unprotected = True

    def __init__(self, app, name, app_name, *args, **kwargs):
        super(AdminComponentMixin, self).__init__(
            name=name, app_name=app_name, *args, **kwargs
        )
        self.app = app

    @property
    def current_app(self):
        return self.app

    def _template_path(self, filename, *args):
        i = super(AdminComponentMixin, self)._template_path(filename, *args)
        for t in i:
            yield t
        yield os.path.join("touchtechnology/admin", filename)

    @never_cache_m
    @staff_login_required_m
    def render(self, request, templates, context, *args, **kwargs):
        """
        If the site has HTTPS enabled, then we should ensure that this view
        is being served securely; otherwise we can carry on as normal.

        Ensure that the AdminComponent will redirect the user if it is deemed
        to be hidden for this tenant. Will have no impact on single-tenant
        installs.
        """
        if hasattr(request, "tenant"):
            admin = context.get("admin")
            if admin._hidden_component(self):
                messages.error(
                    request,
                    'Attempt to access component "%s" '
                    "is not allowed for this tenant." % smart_str(self.verbose_name),
                )
                return HttpResponseRedirect(reverse("admin:index"))

        if SITEMAP_HTTPS_OPTION and not request.is_secure():
            host = request.META.get("HTTP_HOST")
            path = request.META.get("PATH_INFO")
            redirect_to = urlunparse(("https", host, path, "", "", ""))
            return HttpResponseRedirect(redirect_to)

        return super(AdminComponentMixin, self).render(
            request, templates, context, *args, **kwargs
        )

    def reverse(self, name, args=(), kwargs={}, prefix=None):
        if prefix is None:
            prefix = self.app
        return super(AdminComponentMixin, self).reverse(name, args, kwargs, prefix)

    def dropdowns(self):
        """
        When implementing your component you should override this method if the
        component is used to manage a number of models or things. For example,
        a component dealing with Users & Groups should return an iterable of
        tuples:

            (name, path, optional font awesome icon name)

        For example:

            [("Users", "/admin/auth/users/", "user"),
             ("Groups", "/admin/auth/groups/", "group")]
        """
        return ()


class AdminComponent(AdminComponentMixin, Application):
    """
    If an AdminComponent does not share any common functionality with a
    front end application, it can just inherit from this for simplicity
    and backwards compatibility.
    """

    model_form_bases = (BootstrapFormControlMixin, ModelForm)

    def get_urls(self):
        urlpatterns = [
            path("", self.index, name="index"),
        ]
        return urlpatterns

    @property
    def namespace(self):
        """
        If we've subclassed another AdminComponent, chances are we want to
        overload existing templates. We should therefore keep the namespace
        of the oldest parent in the inheritance tree (that is a descendant
        of AdminComponent).
        """
        mro = self.__class__.__mro__
        i = mro.index(AdminComponent)
        return self._get_namespace(mro[i - 1])

    @property
    def template_base(self):
        """
        Our standard template directory for AdminComponent's will be 'admin'
        within the projects normally namespaced template directory.
        """
        return os.path.join(super(AdminComponent, self).template_base, "admin")


class DashboardWidget(object):
    verbose_name = NotImplemented
    template = NotImplemented
    half = True

    @classmethod
    def show(cls):
        return True

    def _get_context(self):
        raise NotImplementedError

    @property
    def context(self):
        return self._get_context()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import touchtechnology.common.db.models
import touchtechnology.common.mixins
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0001_initial"),
        ("contenttypes", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="PythonPackage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("version", models.CharField(max_length=20)),
                ("latest", models.CharField(max_length=20)),
                ("active", models.BooleanField(default=True)),
            ],
            options={"ordering": ("name",),},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="SitemapNode",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "short_title",
                    models.CharField(
                        help_text="This is used in navigation menus instead of the longer title value.",
                        max_length=100,
                        verbose_name="Short title",
                        blank=True,
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "slug_locked",
                    models.BooleanField(default=False, verbose_name="Slug locked"),
                ),
                ("object_id", models.PositiveIntegerField(null=True, blank=True)),
                (
                    "require_https",
                    models.BooleanField(
                        default=False,
                        help_text="Force this to be served via HTTPS.",
                        verbose_name="HTTPS Required",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Set this to 'No' to disable this object and it's children on the site.",
                        verbose_name="Enabled",
                    ),
                ),
                (
                    "hidden_from_navigation",
                    models.BooleanField(
                        default=False,
                        help_text="When set to 'Yes' this object will still be available, but will not appear in menus.",
                        verbose_name="Hide from menus",
                    ),
                ),
                (
                    "hidden_from_sitemap",
                    models.BooleanField(
                        default=False,
                        help_text="When set to 'Yes' this object will not be listed in the auto-generated sitemap.",
                        verbose_name="Hide from sitemap",
                    ),
                ),
                (
                    "hidden_from_robots",
                    models.BooleanField(
                        default=False,
                        help_text="Set this to 'Yes' to prevent search engines from indexing this part of the site.<br />\n<strong>Warning:</strong> this may affect your ranking in search engines.",
                        verbose_name="Hide from spiders",
                    ),
                ),
                ("lft", models.PositiveIntegerField(editable=False, db_index=True)),
                ("rght", models.PositiveIntegerField(editable=False, db_index=True)),
                ("tree_id", models.PositiveIntegerField(editable=False, db_index=True)),
                ("level", models.PositiveIntegerField(editable=False, db_index=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        to="contenttypes.ContentType",
                        null=True,
                        on_delete=models.PROTECT,
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        related_name="children",
                        verbose_name="Parent",
                        blank=True,
                        to="common.SitemapNode",
                        null=True,
                        on_delete=models.PROTECT,
                    ),
                ),
                (
                    "restrict_to_groups",
                    touchtechnology.common.db.models.ManyToManyField(
                        help_text="If you select one or more of these groups your visitors will need to be logged in and a member of an appropriate group to view this part of the site.",
                        to="auth.Group",
                        verbose_name="Restrict to Groups",
                        blank=True,
                    ),
                ),
            ],
            options={"ordering": ("tree_id", "lft"),},
            bases=(touchtechnology.common.mixins.NodeRelationMixin, models.Model),
        ),
    ]



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.hashers import UNUSABLE_PASSWORD_PREFIX


class UserSubclassBackend(ModelBackend):
    def get_user(self, user_id):
        UserModel = get_user_model()
        try:
            return UserModel.objects.exclude(
                password__startswith=UNUSABLE_PASSWORD_PREFIX).get(pk=user_id)
        except UserModel.DoesNotExist:
            return None


class EmailUserSubclassBackend(UserSubclassBackend):
    def authenticate(self, request, username=None, password=None):
        UserModel = get_user_model()
        try:
            # Check if the 'username' is actually an email address
            if username is not None and "@" in username:
                # If there is more than one user account with the same email
                # address on file we'll step through them. Particularly of use
                # when we've switched to email based authentication after
                # previously having used usernames.
                for user in UserModel.objects.filter(email=username):
                    if user.check_password(password):
                        return user
        except UserModel.DoesNotExist:
            pass
        return None



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-09-27 21:08
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('competition', '0024_auto_20170203_0619'),
    ]

    operations = [
        migrations.AlterField(
            model_name='division',
            name='sportingpulse_url',
            field=models.URLField(blank=True, editable=False, max_length=1024, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='division',
            unique_together=set([('slug', 'season'), ('title', 'season')]),
        ),
        migrations.AlterUniqueTogether(
            name='season',
            unique_together=set([('slug', 'competition'), ('title', 'competition')]),
        ),
        migrations.AlterUniqueTogether(
            name='stage',
            unique_together=set([('slug', 'division'), ('title', 'division')]),
        ),
    ]



from django.conf import settings
from django.contrib.sites.models import Site
from django.urls import path
from django.utils.translation import gettext_lazy as _

from touchtechnology.admin.base import AdminComponent
from touchtechnology.common.decorators import superuser_login_required_m


class Settings(AdminComponent):
    verbose_name = _("Settings")
    visible = False

    def __init__(self, app, name="settings", app_name="settings"):
        super(Settings, self).__init__(app, name, app_name)

    @property
    def template_base(self):
        return "touchtechnology/admin/settings"

    def get_urls(self):
        urlpatterns = super(Settings, self).get_urls()
        urlpatterns += [
            path("site/", self.edit_site, name="site"),
        ]
        return urlpatterns

    @superuser_login_required_m
    def index(self, request, *args, **kwargs):
        templates = self.template_path("index.html")
        return self.render(request, templates, kwargs)

    @superuser_login_required_m
    def edit_site(self, request, **extra_context):
        tenant = getattr(request, "tenant", None)
        if tenant is not None:
            return self.generic_edit(
                request,
                tenant.__class__,
                instance=tenant,
                form_class=getattr(tenant, "limited_form_class", None),
                extra_context=extra_context,
            )
        return self.generic_edit(
            request, Site, pk=settings.SITE_ID, extra_context=extra_context
        )



from celery import shared_task

from tournamentcontrol.competition.models import Match, Stage
from tournamentcontrol.competition.utils import generate_scorecards


@shared_task
def generate_pdf_scorecards(
    match_pks, templates, extra_context, stage_pk=None, **kwargs
):
    matches = Match.objects.filter(pk__in=match_pks)
    stage = None
    if stage_pk is not None:
        stage = Stage.objects.get(pk=stage_pk)
    return generate_scorecards(
        matches, templates, "pdf", extra_context, stage, **kwargs
    )



from django.utils.module_loading import import_string


def processor_factory(defs):
    processors = []
    for path, args, kwargs in defs:
        processor = import_string(path)
        processors.append(processor(*args, **kwargs))
    return processors



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("competition", "0014_match_play_at_blank"),
    ]

    operations = [
        migrations.AddField(
            model_name="person",
            name="user",
            field=models.OneToOneField(
                null=True,
                blank=True,
                to=settings.AUTH_USER_MODEL,
                on_delete=models.deletion.PROTECT,
            ),
            preserve_default=True,
        ),
    ]



import datetime
import functools
from functools import reduce
from operator import add, or_

from dateutil.parser import parse
from django import forms
from django.db.models import Q
from django.http import HttpResponse, HttpResponseRedirect
from django.utils.translation import gettext_lazy as _
from first import first
from formtools.wizard.views import SessionWizardView

from touchtechnology.common.forms.fields import (
    ModelChoiceField,
    ModelMultipleChoiceField,
)
from tournamentcontrol.competition.models import Season
from tournamentcontrol.competition.tasks import generate_pdf_scorecards
from tournamentcontrol.competition.utils import generate_scorecards

##############################################################################
# Scorecard Report
##############################################################################


def combine_date_time_tuple(t):
    return datetime.datetime.combine(*t)


def q_date_from_datetime(dt):
    return Q(date=dt.date())


def q_date_time_from_datetime(dt):
    return Q(date=dt.date(), time=dt.time())


no_date_or_time = Q(date__isnull=True) | Q(time__isnull=True)


class SeasonForm(forms.Form):
    """
    Step 1. Select which season you want to generate a report of.
    """

    def __init__(self, *args, **kwargs):
        super(SeasonForm, self).__init__(*args, **kwargs)

        seasons = (
            Season.objects.exclude(complete=True)
            .select_related("competition")
            .order_by("-start_date")
        )
        self.fields["season"] = ModelChoiceField(
            queryset=seasons,
            label_from_instance=lambda s: "{0} {1}".format(
                s.competition.title, s.title
            ),
        )


class FilterForm(forms.Form):
    """
    Step 2. Provide fields to allow the admin to filter the matches for
    inclusion in the report.
    """

    def __init__(self, season, *a, **kw):
        super(FilterForm, self).__init__(*a, **kw)

        # given a Season instance get all the matches which should have a
        # scorecard produced.
        self.matches = season.matches.exclude(no_date_or_time)

        # allow filtering by division
        self.fields["division"] = ModelMultipleChoiceField(
            queryset=season.divisions.all(), label_from_instance="title", required=False
        )

        # allow filtering by date
        dates = self.matches.order_by("date").values_list("date", flat=True).distinct()

        date_choices = zip(dates, dates)
        self.fields["dates"] = forms.MultipleChoiceField(
            choices=date_choices, widget=forms.CheckboxSelectMultiple, required=False
        )

        # allow filtering by timeslot
        timeslots = (
            self.matches.order_by("date", "time").values_list("date", "time").distinct()
        )

        datetimes = [combine_date_time_tuple(t) for t in timeslots]
        datetime_choices = zip(datetimes, datetimes)
        self.fields["timeslots"] = forms.MultipleChoiceField(
            choices=datetime_choices,
            widget=forms.CheckboxSelectMultiple,
            required=False,
        )

        # allow configurable templates to be rendered
        template_choices = (
            ("", _("Choose a report template...")),
            ("scorecards.html", "Scorecards"),
            ("signon.html", "Sign-on Sheet"),
            ("mvp.html", "Most Valuable Player"),
        )

        self.fields["template"] = forms.ChoiceField(
            choices=template_choices, label=_("Template")
        )

        output_choices = (
            ("", _("Choose an output format...")),
            ("pdf", _("PDF")),
            ("html", _("HTML")),
        )
        self.fields["format"] = forms.ChoiceField(
            choices=output_choices, label=_("Output as")
        )

    def clean(self):
        data = self.cleaned_data.copy()
        query_filter = Q()

        dates = self.cleaned_data.get("dates")
        timeslots = self.cleaned_data.get("timeslots")

        if dates:
            date_list = [q_date_from_datetime(parse(d)) for d in dates]
            query_filter |= functools.reduce(or_, date_list)

        if timeslots:
            timeslot_list = [q_date_time_from_datetime(parse(t)) for t in timeslots]
            query_filter |= functools.reduce(or_, timeslot_list)

        divisions = self.cleaned_data.get("division")

        if divisions:
            query_filter &= Q(stage__division__in=divisions)

        matches = self.matches.filter(query_filter)
        if not matches.count():
            raise forms.ValidationError(
                _("There are no matches that satisfy " "your criteria.")
            )

        data["matches"] = matches

        return data


class ScorecardWizardBase(SessionWizardView):
    def get_context_data(self, form, **kwargs):
        context = super(ScorecardWizardBase, self).get_context_data(form, **kwargs)
        context.update(self.extra_context)
        return context

    def get_form_kwargs(self, step):
        # pass the cleaned_data from the previous step into the current step
        return self.get_cleaned_data_for_step(self.get_prev_step(step=step)) or {}

    def get_template_names(self):
        return self.app.template_path("scorecards/report.html")

    def done(self, form_list, **kwargs):
        data = self.get_all_cleaned_data()
        format = data.get("format")
        season = data.pop("season")
        matches = data.pop("matches")
        template = data.pop("template")

        extra_context = {
            "competition": season.competition,
            "season": season,
        }

        templates = self.app.template_path(
            template, season.competition.slug, season.slug
        )

        if format == "pdf":
            kw = {
                "matches": matches,
                "templates": templates,
                "extra_context": extra_context,
            }
            if hasattr(self.request, "tenant"):
                kw["_schema_name"] = self.request.tenant.schema_name
                kw["base_url"] = self.request.build_absolute_uri("/")

            result = generate_pdf_scorecards.delay(**kw)

            reverse_kwargs = {
                "competition_id": season.competition_id,
                "season_id": season.pk,
                "result_id": result.id,
            }

            redirect_to = self.app.reverse(
                "competition:season:scorecards-async", kwargs=reverse_kwargs
            )
            response = HttpResponseRedirect(redirect_to)
        else:
            output = generate_scorecards(matches, templates, extra_context)
            response = HttpResponse(output)

        return response


def scorecardwizard_factory(**kwargs):
    return type("ScorecardWizard", (ScorecardWizardBase,), kwargs)


##############################################################################
# Draw Generation
##############################################################################


class DrawGenerationWizard(SessionWizardView):

    extra_context = {}
    redirect_to = None
    stage = None
    template_name = "tournamentcontrol/competition/admin/draw_generation/wizard.html"

    def get_form_initial(self, step):
        if step == "0":
            if self.stage.pools.count():
                queryset = self.stage.pools.all()
            else:
                queryset = self.stage.division.stages.filter(pk=self.stage.pk)
            return queryset

    def get_form_kwargs(self, step):
        if step == "1":
            data = self.get_cleaned_data_for_step("0")
            matches = functools.reduce(add, [d["matches"] for d in data])
            return {"queryset": matches}
        return {}

    def get_context_data(self, form, **kwargs):
        context = super(DrawGenerationWizard, self).get_context_data(
            form=form, **kwargs
        )
        context.update(self.extra_context)
        context.update(
            {
                "model": self.stage.matches.none(),
            }
        )
        return context

    def done(self, form_list, **kwargs):
        first(reversed(form_list)).save()
        return HttpResponseRedirect(self.redirect_to)



from __future__ import unicode_literals

import collections
import datetime
import json
import logging

from dateutil.parser import parse
from dateutil.rrule import DAILY, WEEKLY
from django import forms
from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.forms import array as PGA
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db.models import Q
from django.forms import BooleanField as BooleanChoiceField
from django.forms.formsets import (
    DELETION_FIELD_NAME,
    INITIAL_FORM_COUNT,
    MAX_NUM_FORM_COUNT,
    TOTAL_FORM_COUNT,
    ManagementForm,
    formset_factory,
)
from django.forms.models import (
    BaseInlineFormSet,
    BaseModelFormSet,
    inlineformset_factory,
    modelformset_factory,
)
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.translation import gettext_lazy as _, ngettext
from first import first
from modelforms.forms import ModelForm
from pyparsing import ParseException

from touchtechnology.common.forms.fields import (
    ModelChoiceField,
    ModelMultipleChoiceField,
    SelectDateField,
)
from touchtechnology.common.forms.mixins import (
    BootstrapFormControlMixin,
    SuperUserSlugMixin,
    UserMixin,
)
from touchtechnology.common.forms.tz import timezone_choice
from touchtechnology.common.forms.widgets import (
    SelectDateTimeWidget as SelectDateTimeWidgetBase,
)
from touchtechnology.content.forms import PlaceholderConfigurationBase
from tournamentcontrol.competition.calc import BonusPointCalculator, Calculator
from tournamentcontrol.competition.draw import seeded_tournament
from tournamentcontrol.competition.fields import URLField
from tournamentcontrol.competition.models import (
    ByeTeam,
    Club,
    ClubAssociation,
    ClubRole,
    Competition,
    Division,
    DivisionExclusionDate,
    DrawFormat,
    Ground,
    LadderEntry,
    Match,
    Person,
    Place,
    Season,
    SeasonAssociation,
    SeasonExclusionDate,
    SeasonMatchTime,
    SimpleScoreMatchStatistic,
    Stage,
    StageGroup,
    Team,
    TeamAssociation,
    TeamRole,
    UndecidedTeam,
    Venue,
    stage_group_position_re,
)
from tournamentcontrol.competition.signals.custom import score_updated
from tournamentcontrol.competition.utils import (
    FauxQueryset,
    legitimate_bye_match,
    match_unplayed,
    time_choice,
)

logger = logging.getLogger(__name__)

EMPTY = [("", "---")]
SCORE_FIELDS = set(["home_team_score", "away_team_score"])

valid_ladder_identifiers = collections.OrderedDict(
    (
        ("win", "Win"),
        ("draw", "Draw"),
        ("loss", "Loss"),
        ("bye", "Bye"),
        ("forfeit_for", "Win by forfeit"),
        ("forfeit_against", "Loss by forfeit"),
        # Other potential identifiers, but they don't really make sense as
        # variables for use in generating a ladder formula.
        #
        # 'score_for', 'score_against', 'played'
    )
)


def ladder_points_widget(name, **attrs):
    defaults = {"placeholder": valid_ladder_identifiers[name].lower()}
    defaults.update(attrs)
    return forms.TextInput(attrs=defaults)


class LadderPointsWidget(forms.MultiWidget):
    def __init__(self, attrs=None):
        self.attrs = attrs or {}

        widgets = tuple(
            [ladder_points_widget(n, **self.attrs) for n in valid_ladder_identifiers]
        )
        super(LadderPointsWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if not value:
            return ""
        values = []
        for i in valid_ladder_identifiers:
            ladder_entry = LadderEntry(**{i: 1})
            calc = Calculator(ladder_entry)
            calc.parse(value)
            values.append(calc.evaluate() or None)
        return values

    def format_output(self, rendered_widgets):
        output = ""
        for label, widget in zip(valid_ladder_identifiers.values(), rendered_widgets):
            output += """
                <div class="field_wrapper">
                    <label class="field_name">{label}</label>
                    <div class="field text_input short">
                        {widget}
                    </div>
                </div>
            """.format(
                label=label, widget=widget
            )
        return output


class MatchPlayedWidget(forms.widgets.Select):
    """
    Custom widget to show a Yes/No choice field for whether a player
    took part in a match or not.

    Based on django.forms.widget.NullBooleanSelect
    """

    def __init__(self, attrs=None):
        choices = (("1", _("Yes")), ("0", _("No")))
        super(MatchPlayedWidget, self).__init__(attrs, choices)

    def value_from_datadict(self, data, files, name):
        value = data.get(name, None)
        return {"1": 1, "0": 0, True: 1, False: 0, "True": 1, "False": 0}.get(
            value, None
        )


class LadderPointsField(forms.MultiValueField):
    def __init__(self, max_length=None, *args, **kwargs):
        fields = tuple(
            map(
                lambda field: forms.IntegerField(required=False, initial=0),
                valid_ladder_identifiers,
            )
        )
        kwargs["widget"] = LadderPointsWidget()
        super(LadderPointsField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        parts = [
            "{}*{}".format(*each)
            for each in zip(data_list, valid_ladder_identifiers)
            if each[0]
        ]
        return " + ".join(parts)


class SelectDateTimeWidget(SelectDateTimeWidgetBase):
    def decompress(self, value):
        """
        Value is serialized as string, so we use the dateutil.parser
        functionality to restore it to primitive datetime instance before then
        calling the parent class decompress function.
        """
        if value:
            value = parse(value)
        return super(SelectDateTimeWidget, self).decompress(value)


class ConstructFormMixin(object):
    """
    When a custom FormSet requires the ability to pass keyword arguments to a
    child form, we can simply define the ``get_defaults`` method to return a
    dictionary to be passed through to the form constructor.
    """

    def get_defaults(self):
        return {}

    def _construct_form(self, i, **kwargs):
        defaults = self.get_defaults()
        defaults.update(kwargs)
        return super(ConstructFormMixin, self)._construct_form(i, **defaults)

    @property
    def empty_form(self):
        defaults = self.get_defaults()
        form = self.form(
            auto_id=self.auto_id,
            prefix=self.add_prefix("__prefix__"),
            empty_permitted=True,
            **defaults
        )
        self.add_fields(form, None)
        return form


class ConfigurationForm(PlaceholderConfigurationBase):
    """
    Configuration of the specific division is performed by slug rather than
    primary key so that we can simply pass through fake url components to our
    underlying views.
    """

    def __init__(self, *args, **kwargs):
        super(ConfigurationForm, self).__init__(*args, **kwargs)

        competitions = Competition.objects.values_list("slug", "title")
        self.fields["competition"] = forms.ChoiceField(
            choices=EMPTY + list(competitions), required=False, label=_("Competition")
        )
        self.fields["competition"].widget.attrs.update({"class": "form-control"})

        seasons = Season.objects.values_list("slug", "title").distinct()
        self.fields["season"] = forms.ChoiceField(
            choices=EMPTY + list(seasons), required=False, label=_("Season")
        )
        self.fields["season"].widget.attrs.update({"class": "form-control"})

    def clean_season(self):
        competition_slug = self.cleaned_data.get("competition")
        season_slug = self.cleaned_data.get("season")
        if season_slug and not competition_slug:
            raise forms.ValidationError(
                _("You can't select a season without a competition.")
            )
        elif (
            season_slug
            and not Season.objects.filter(
                competition__slug=competition_slug, slug=season_slug
            ).exists()
        ):
            raise forms.ValidationError(_("Invalid season for this competition."))
        return season_slug


class MultiConfigurationForm(PlaceholderConfigurationBase):
    def __init__(self, *args, **kwargs):
        super(MultiConfigurationForm, self).__init__(*args, **kwargs)

        competitions = Competition.objects.values_list("slug", "title")
        self.fields["competition"] = forms.MultipleChoiceField(
            choices=list(competitions),
            required=True,
            label=_("Competition"),
            widget=forms.CheckboxSelectMultiple,
        )


class RankingConfigurationForm(PlaceholderConfigurationBase):
    def __init__(self, *args, **kwargs):
        super(RankingConfigurationForm, self).__init__(*args, **kwargs)
        self.fields["start"] = forms.CharField(required=False)
        self.fields["decay"] = forms.CharField(required=False)


class PersonEditForm(BootstrapFormControlMixin, ModelForm):
    class Meta:
        model = Person
        fields = (
            "first_name",
            "last_name",
            "gender",
            "date_of_birth",
            "email",
            "home_phone",
            "work_phone",
            "mobile_phone",
            "user",
        )


class PersonMergeForm(PersonEditForm):

    merge = ModelMultipleChoiceField(
        label=_("Duplicates"),
        queryset=Person.objects.all(),
        label_from_instance="get_full_name",
        widget=forms.SelectMultiple,
        help_text=_(
            "Select the duplicate members that are to be merged "
            "into this retained record."
        ),
    )
    keep_old = forms.TypedChoiceField(
        choices=(("true", "Yes"), ("false", "No")), coerce=json.loads, initial="false"
    )

    def __init__(self, *args, **kwargs):
        super(PersonMergeForm, self).__init__(*args, **kwargs)
        self.fields["merge"].queryset = self.instance.club.members.exclude(
            pk=self.instance.pk
        )

    def save(self, *args, **kwargs):
        from tournamentcontrol.competition.merge import merge_model_objects

        keep = super(PersonMergeForm, self).save(*args, **kwargs)
        return merge_model_objects(
            keep, list(self.cleaned_data["merge"]), self.cleaned_data["keep_old"]
        )


class CompetitionForm(SuperUserSlugMixin, ModelForm):
    class Meta:
        model = Competition
        fields = (
            "title",
            "short_title",
            "rank_importance",
            "enabled",
            "copy",
            "slug",
            "slug_locked",
            "clubs",
        )
        labels = {
            "copy": _("Description"),
        }


class TimezoneMixin(object):
    """
    Mixin to remove the timezone field if we are not operating in a timezone
    aware state.
    """

    def __init__(self, *args, **kwargs):
        super(TimezoneMixin, self).__init__(*args, **kwargs)
        if not settings.USE_TZ:
            self.fields.pop("timezone", None)


class SeasonForm(
    SuperUserSlugMixin, TimezoneMixin, BootstrapFormControlMixin, ModelForm
):
    class Meta:
        model = Season
        fields = (
            "title",
            "short_title",
            "copy",
            "hashtag",
            "enabled",
            "timezone",
            "start_date",
            "mode",
            "forfeit_notifications",
            "complete",
            "statistics",
            "mvp_results_public",
            "slug",
            "slug_locked",
        )
        labels = {
            "copy": _("Notes (Public)"),
        }
        help_texts = {
            "copy": _("Optional. Will be displayed in the front end if provided."),
        }


class VenueForm(SuperUserSlugMixin, TimezoneMixin, ModelForm):
    class Meta:
        model = Venue
        fields = (
            "title",
            "short_title",
            "abbreviation",
            "timezone",
            "latlng",
            "slug",
            "slug_locked",
        )
        labels = {
            "latlng": _("Map"),
        }


class GroundForm(SuperUserSlugMixin, TimezoneMixin, ModelForm):
    class Meta:
        model = Ground
        fields = (
            "title",
            "short_title",
            "abbreviation",
            "timezone",
            "latlng",
            "slug",
            "slug_locked",
        )
        labels = {
            "latlng": _("Map"),
        }


BaseGroundFormSet = inlineformset_factory(
    Venue, Ground, fk_name="venue", fields=("title", "abbreviation"), extra=0
)


class GroundFormSet(BaseGroundFormSet):
    def _construct_form(self, i, **kwargs):
        if i >= self.initial_form_count():
            kwargs["instance"] = self.model(order=i + 1)
        return super(GroundFormSet, self)._construct_form(i, **kwargs)


class DivisionForm(SuperUserSlugMixin, ModelForm):
    def __init__(self, *args, **kwargs):
        super(DivisionForm, self).__init__(*args, **kwargs)
        if self.instance.season.mode != DAILY:
            self.fields.pop("games_per_day", None)

    class Meta:
        model = Division
        fields = (
            "title",
            "short_title",
            "rank_division",
            "copy",
            "draft",
            "points_formula",
            "bonus_points_formula",
            "games_per_day",
            "forfeit_for_score",
            "forfeit_against_score",
            "include_forfeits_in_played",
            "slug",
            "slug_locked",
        )
        labels = {
            "copy": _("Notes (Public)"),
            "forfeit_for_score": _("Forfeit win score"),
            "forfeit_against_score": _("Forfeit loss score"),
            "include_forfeits_in_played": _("Add forfeits to played"),
        }
        help_texts = {
            "copy": _("Optional. Will be displayed in the front end if provided."),
        }
        widgets = {
            "bonus_points_formula": forms.TextInput,
        }

    def _clean_formula(self, field_name, calculator_class):
        """
        Generic clean function for both the formula fields.
        """
        fake = LadderEntry()
        formula = self.cleaned_data.get(field_name)
        parser = calculator_class(fake)
        try:
            parser.parse(formula)
        except ParseException:
            raise forms.ValidationError(_("Syntax of this points formula is invalid."))
        return formula

    def clean_points_formula(self):
        return self._clean_formula("points_formula", Calculator)

    def clean_bonus_points_formula(self):
        return self._clean_formula("bonus_points_formula", BonusPointCalculator)


class StageForm(SuperUserSlugMixin, ModelForm):
    def __init__(self, *args, **kwargs):
        super(StageForm, self).__init__(*args, **kwargs)
        if self.instance.order <= 1:
            self.fields.pop("follows")
        else:
            self_or_higher_order = Q(pk=self.instance.pk) | Q(
                order__gt=self.instance.order
            )
            self.fields["follows"].queryset = self.instance.division.stages.exclude(
                self_or_higher_order
            )
            self.fields["follows"].empty_label = _("Default")

    class Meta:
        model = Stage
        fields = (
            "title",
            "short_title",
            "follows",
            "rank_importance",
            "keep_ladder",
            "scale_group_points",
            "carry_ladder",
            "keep_mvp",
            "slug",
            "slug_locked",
        )


class StageGroupForm(SuperUserSlugMixin, ModelForm):

    teams = ModelMultipleChoiceField(queryset=None)

    def __init__(self, *args, **kwargs):
        super(StageGroupForm, self).__init__(*args, **kwargs)

        # build a queryset of teams in the division that are either assigned
        # to this pool or not assigned to any other pools in this stage
        other_pools = self.instance.stage.pools.exclude(pk=self.instance.pk)

        if self.instance.stage.order > 1 and not self.instance.teams.count():
            self._undecided = True
            queryset = self.instance.stage.undecided_teams.exclude(
                stage_group__in=other_pools
            )
            label_from_instance = "title"
            initial = self.instance.undecided_teams.all()
        else:
            self._undecided = False
            queryset = self.instance.stage.division.teams.exclude(
                stage_group__in=other_pools
            )
            label_from_instance = "title"
            initial = self.instance.teams.all()

        self.fields["teams"] = ModelMultipleChoiceField(
            queryset=queryset.order_by(*queryset.model._meta.ordering),
            initial=initial,
            required=False,
            label_from_instance=label_from_instance,
            help_text=_(
                "Teams can only belong to one pool per stage. Once "
                "selected in any pool a team will no longer appear "
                "for selection here."
            ),
        )

        if self.instance.matches.exists():
            self.fields.pop("teams", None)

    def save(self, *args, **kwargs):
        if self.instance.pk and "teams" in self.fields:
            if self._undecided:
                self.instance.undecided_teams = self.cleaned_data.get("teams")
            else:
                self.instance.teams = self.cleaned_data.get("teams")
        return super(StageGroupForm, self).save(*args, **kwargs)

    class Meta:
        model = StageGroup
        fields = (
            "title",
            "short_title",
            "rank_importance",
            "carry_ladder",
            "teams",
            "slug",
            "slug_locked",
        )


class StageGroupFormSetForm(ModelForm):
    def __init__(self, stage, *args, **kwargs):
        super(StageGroupFormSetForm, self).__init__(*args, **kwargs)

    class Meta:
        model = StageGroup
        fields = ("title",)


BaseStageGroupFormSet = inlineformset_factory(
    Stage, StageGroup, form=StageGroupFormSetForm, extra=0
)


class StageGroupFormSet(ConstructFormMixin, BaseStageGroupFormSet):
    def __init__(self, stage, *args, **kwargs):
        self.stage = stage
        super(StageGroupFormSet, self).__init__(*args, **kwargs)

    def get_defaults(self):
        return {"stage": self.stage}

    def _construct_form(self, i, **kwargs):
        if i >= self.initial_form_count():
            kwargs["instance"] = self.model(stage=self.stage, order=i + 1)
        return super(StageGroupFormSet, self)._construct_form(i, **kwargs)


class UndecidedTeamForm(UserMixin, ModelForm):
    class Meta:
        model = UndecidedTeam
        fields = (
            "formula",
            "label",
        )
        help_texts = {
            "formula": _(
                "If left blank you will need to manually progress a team and "
                "the <em>label</em> below will be required for display in "
                "the draw."
            ),
            "label": _(
                "Required if formula is left blank. If formula is set this "
                "value will be ignored and not saved."
            ),
        }

    def clean_formula(self):
        # TODO move to general clean and insert error back on field so that
        # validation of `label` is not impacted by this.
        formula = self.cleaned_data.get("formula").upper()
        if formula and not stage_group_position_re.match(formula):
            raise forms.ValidationError(_("This is not a valid team formula."))
        return formula

    def clean_label(self):
        formula = self.cleaned_data.get("formula")
        label = self.cleaned_data.get("label")
        if formula:
            label = ""
        elif not label:
            raise forms.ValidationError(
                _("You must specify a label when formula is blank.")
            )
        return label


class TeamForm(SuperUserSlugMixin, ModelForm):
    def __init__(self, division, *args, **kwargs):
        super(TeamForm, self).__init__(*args, **kwargs)
        if not division.season.competition.clubs.count():
            self.fields.pop("club")
        else:
            self.fields["club"].queryset = division.season.competition.clubs
            self.fields["title"].required = False

        timeslot_fields = ("timeslots_after", "timeslots_before")

        # If the season has timeslot rules, substitute the default SelectTime
        # field for a drop-down list of timeslots.
        if self.instance.division.season.mode != WEEKLY:
            for field in timeslot_fields:
                self.fields.pop(field, None)
                setattr(self.instance, field, None)
        elif self.instance.division.season.timeslots.count():
            timeslots = self.instance.division.season.get_timeslots()
            choices = [("", "")] + [time_choice(t) for t in timeslots]

            for field in timeslot_fields:
                self.fields[field] = forms.TimeField(
                    required=False,
                    widget=forms.Select,
                    label=self.fields[field].label,
                    help_text=self.fields[field].help_text,
                )
                self.fields[field].widget.attrs = {"class": "form-control"}
                self.fields[field].widget.choices = choices

                # Ensure that the currently set time is in the list if not in
                # the rruleset defined for the season.
                value = getattr(self.instance, field)
                if value:
                    initial = time_choice(value)
                    if initial not in choices:
                        self.fields[field].widget.choices.append(initial)
                        self.fields[field].widget.choices.sort()

        # Update the set of teams which we don't want this team to clash with
        # when allocating matches. Exclude self, and order by team name.
        queryset = self.fields["team_clashes"].queryset.filter(
            division__season=self.instance.division.season
        )
        # Also exclude teams in the same division, you have to play them so
        # this will result in a clash that time.
        queryset = queryset.exclude(division=self.instance.division)
        if self.instance.pk:
            queryset = queryset.exclude(pk=self.instance.pk)
        self.fields["team_clashes"].queryset = queryset.select_related(
            "division"
        ).order_by("title", "division")
        if not queryset.count():
            self.fields.pop("team_clashes", None)

    class Meta:
        model = Team
        fields = (
            "club",
            "title",
            "short_title",
            "rank_division",
            "copy",
            "names_locked",
            "slug",
            "slug_locked",
            "timeslots_after",
            "timeslots_before",
            "team_clashes",
        )
        labels = {
            "title": _("Name"),
            "short_title": _("Abbreviation"),
            "names_locked": _("Name is locked"),
            "copy": _("Notes (Public)"),
        }
        help_texts = {
            "copy": _("Optional. Will be displayed in the front end if provided."),
            "short_title": _(
                "The abbreviated version will be used in draws " "to save space."
            ),
        }

    class Media:
        js = ("tournamentcontrol/competition/js/team.js",)

    def clean_title(self):
        club = self.cleaned_data.get("club")
        title = self.cleaned_data.get("title")
        if not title and club:
            self.data["title"] = club.title
            return club.title
        return title


class DrawFormatForm(BootstrapFormControlMixin, ModelForm):
    def clean_text(self):
        from tournamentcontrol.competition.draw import DrawGenerator

        text = self.cleaned_data.get("text").strip()
        try:
            DrawGenerator.validate(text)
        except ValueError as e:
            raise forms.ValidationError(e.message)
        return text

    class Meta:
        model = DrawFormat
        fields = ("name", "text", "teams", "is_final")


class BaseMatchFormMixin(BootstrapFormControlMixin):
    def __init__(self, *args, **kwargs):
        super(BaseMatchFormMixin, self).__init__(*args, **kwargs)

        # set the queryset of the `home_team` and `away_team` fields
        team_ids = self.instance.stage.division.teams.values_list("id", flat=True)
        undecided_team_ids = self.instance.stage.undecided_teams.values_list(
            "id", flat=True
        )

        for prefix in ("home", "away"):
            for suffix, ids in (("", team_ids), ("_undecided", undecided_team_ids)):
                field = prefix + "_team" + suffix
                if field in self.fields:
                    self.fields[field].queryset = self.fields[field].queryset.filter(
                        id__in=ids
                    )

        # If the season has timeslot rules, substitute the default SelectTime
        # field for a drop-down list of timeslots.
        if self.instance.stage.division.season.timeslots.count():
            timeslots = self.instance.stage.division.season.get_timeslots(
                self.instance.date
            )

            if self.instance.home_team and self.instance.home_team.timeslots_after:
                timeslots = [
                    timeslot
                    for timeslot in timeslots
                    if timeslot >= self.instance.home_team.timeslots_after
                ]

            if self.instance.away_team and self.instance.away_team.timeslots_after:
                timeslots = [
                    timeslot
                    for timeslot in timeslots
                    if timeslot >= self.instance.away_team.timeslots_after
                ]

            if self.instance.home_team and self.instance.home_team.timeslots_before:
                timeslots = [
                    timeslot
                    for timeslot in timeslots
                    if timeslot <= self.instance.home_team.timeslots_before
                ]

            if self.instance.away_team and self.instance.away_team.timeslots_before:
                timeslots = [
                    timeslot
                    for timeslot in timeslots
                    if timeslot <= self.instance.away_team.timeslots_before
                ]

            choices = [("", "")] + [time_choice(timeslot) for timeslot in timeslots]

            # Ensure that the currently set time is in the list if not in the
            # rruleset defined for the season.
            if self.instance.time:
                initial = time_choice(self.instance.time)
                if initial not in choices:
                    choices.append(initial)
                choices.sort()

            # Only replace the widget if we were able to resolve at least 1
            # timeslot from the season -- otherwise it will be impossible for
            # the user to set a value at all.
            if timeslots and "time" in self.fields:  # Check if in higher up?
                self.fields["time"] = forms.TimeField(
                    required=False, widget=forms.Select
                )
                self.fields["time"].widget.choices = choices
                self.fields["time"].widget.attrs.setdefault("class", "form-control")


def _match_edit_form_formfield_callback(field, **kwargs):
    """
    Customise fields in the :class:`MatchEditForm`

    Set the size of the videos array with
    `settings.TOURNAMENTCONTROL_COMPETITION_VIDEOS_ARRAY_SIZE` (default is 5)
    """
    if isinstance(field, ArrayField):
        size = getattr(settings, "TOURNAMENTCONTROL_COMPETITION_VIDEOS_ARRAY_SIZE", 5)
        formfield = PGA.SplitArrayField(
            URLField(required=False),
            size,
            remove_trailing_nulls=True,
            required=False,
            label=_("Video URLs"),
            help_text=_(
                "You can provide up to %(size)d videos per match. "
                "YouTube and Vimeo links are supported."
            )
            % {
                "size": size,
            },
        )
        return formfield
    return field.formfield(**kwargs)


class MatchEditForm(BaseMatchFormMixin, ModelForm):
    """
    Use this form to make sure that the data validates:

     * that if the `stage_group` is set, it is a group of the `division`
     * that both the `home_team` and `away_team` are in the `division`
    """

    formfield_callback = _match_edit_form_formfield_callback

    def __init__(self, *args, **kwargs):
        super(MatchEditForm, self).__init__(*args, **kwargs)

        # restrict the list of referees to those registered this season
        if "referees" in self.fields:
            self.fields[
                "referees"
            ].queryset = self.instance.stage.division.season.referees.all()

        # remove `stage_group` field if the `division` has no children
        if not self.instance.stage.pools.count():
            self.fields.pop("stage_group", None)
        elif "stage_group" in self.fields:
            self.fields["stage_group"].queryset = self.instance.stage.pools.all()

        # restrict the team choices if we have pools set
        if self.instance.stage_group:
            self.fields["home_team"].queryset = self.instance.stage_group.teams.all()
            self.fields["home_team"].empty_label = ""
            if self.instance.home_team:
                self.fields["home_team"].queryset |= Team.objects.filter(
                    pk=self.instance.home_team.pk
                )
            self.fields["away_team"].queryset = self.instance.stage_group.teams.all()
            self.fields["away_team"].empty_label = ""
            if self.instance.away_team:
                self.fields["away_team"].queryset |= Team.objects.filter(
                    pk=self.instance.away_team.pk
                )

        # remove team fields depending on if we are editing a match with known
        # teams, undecided teams, or teams that still need to be evaluated.
        if not self.instance.home_team and self.instance.stage.undecided_teams.count():
            self.fields.pop("home_team", None)

        if (
            not self.instance.home_team_undecided
            and not self.instance.stage.undecided_teams.count()
        ):
            self.fields.pop("home_team_undecided", None)
            try:
                empty_label = self.instance.get_home_team()["title"]
            except TypeError:
                empty_label = "---"
            self.fields["home_team"].empty_label = empty_label

        if not self.instance.away_team and self.instance.stage.undecided_teams.count():
            self.fields.pop("away_team", None)

        if (
            not self.instance.away_team_undecided
            and not self.instance.stage.undecided_teams.count()
        ):
            self.fields.pop("away_team_undecided", None)
            try:
                empty_label = self.instance.get_away_team()["title"]
            except TypeError:
                empty_label = "---"
            self.fields["away_team"].empty_label = empty_label

    def clean_home_team(self):
        pool = self.cleaned_data.get("stage_group")
        team = self.cleaned_data.get("home_team")
        if (
            pool
            and team
            and team not in pool.teams.all()
            and team != self.instance.home_team
        ):
            raise forms.ValidationError(
                _('Team "%(team)s" is not in pool "%(pool)s".')
                % {"team": team.title, "pool": pool.title}
            )
        if (
            team
            and team not in self.instance.stage.division.teams.all()
            and team != self.instance.home_team
        ):
            raise forms.ValidationError(_("This team is not in this division."))
        return team

    def clean_away_team(self):
        pool = self.cleaned_data.get("stage_group")
        team = self.cleaned_data.get("away_team")
        if (
            pool
            and team
            and team not in pool.teams.all()
            and team != self.instance.away_team
        ):
            raise forms.ValidationError(
                _('Team "%(team)s" is not in pool "%(pool)s".')
                % {"team": team.title, "pool": pool.title}
            )
        if (
            team
            and team not in self.instance.stage.division.teams.all()
            and team != self.instance.away_team
        ):
            raise forms.ValidationError(_("This team is not in this division."))
        if team and team == self.cleaned_data.get("home_team"):
            raise forms.ValidationError(
                _("Teams cannot be scheduled to play against itself.")
            )
        return team

    def clean_home_team_undecided(self):
        pool = self.cleaned_data.get("stage_group")
        team = self.cleaned_data.get("home_team_undecided")
        if pool and team and team not in pool.undecided_teams.all():
            raise forms.ValidationError(
                _('Team "%(team)s" is not in pool "%(pool)s".')
                % {"team": team.title, "pool": pool.title}
            )
        if team and team not in self.instance.stage.undecided_teams.all():
            raise forms.ValidationError(_("This team is not in this division."))
        return team

    def clean_away_team_undecided(self):
        pool = self.cleaned_data.get("stage_group")
        team = self.cleaned_data.get("away_team_undecided")
        if pool and team and team not in pool.undecided_teams.all():
            raise forms.ValidationError(
                _('Team "%(team)s" is not in pool "%(pool)s".')
                % {"team": team.title, "pool": pool.title}
            )
        if team and team not in self.instance.stage.undecided_teams.all():
            raise forms.ValidationError(_("This team is not in this division."))
        if team and team == self.cleaned_data.get("home_team_undecided"):
            raise forms.ValidationError(
                _("Teams cannot be scheduled to play against itself.")
            )
        return team

    def clean_videos(self):
        videos = self.cleaned_data.get("videos")
        if any(videos):
            return videos

    class Meta:
        model = Match
        fields = (
            "stage_group",
            "home_team",
            "away_team",
            "home_team_undecided",
            "away_team_undecided",
            "referees",
            "label",
            "round",
            "date",
            "include_in_ladder",
            "videos",
        )
        labels = {
            "home_team_undecided": _("Home team"),
            "away_team_undecided": _("Away team"),
        }


class DrawGenerationMatchForm(MatchEditForm):
    def has_changed(self):
        # This seems like a hack to make DrawGenerationMatchFormSet work on
        # Django 1.7, see django/forms/models.py:758 and save_new_objects
        # implementation.
        return True

    class Meta(MatchEditForm.Meta):
        fields = (
            "round",
            "home_team",
            "away_team",
            "home_team_undecided",
            "away_team_undecided",
            "date",
        )


BaseDrawGenerationMatchFormSet = modelformset_factory(
    Match, form=DrawGenerationMatchForm, extra=0, can_delete=True
)


class DrawGenerationMatchFormSet(BaseDrawGenerationMatchFormSet):
    def _construct_form(self, i, **kwargs):
        if i < self.initial_form_count() and not kwargs.get("instance"):
            kwargs["instance"] = self.get_queryset()[i]
        return super(BaseModelFormSet, self)._construct_form(i, **kwargs)

    def add_fields(self, form, index):
        """A hook for adding extra fields on to each form instance."""
        super(DrawGenerationMatchFormSet, self).add_fields(form, index)
        if self.can_delete:
            form.fields[DELETION_FIELD_NAME] = BooleanChoiceField(
                label=_("Skip"), required=False
            )

    def save(self, *args, **kwargs):
        matches = []
        for form in self.forms:
            if form.instance.home_team_eval_related is not None:
                form.instance.home_team_eval_related = Match.objects.get(
                    uuid=form.instance.home_team_eval_related.uuid
                )
            if form.instance.away_team_eval_related is not None:
                form.instance.away_team_eval_related = Match.objects.get(
                    uuid=form.instance.away_team_eval_related.uuid
                )
            matches.append(form.save())
        return matches


class MatchResultForm(BootstrapFormControlMixin, ModelForm):
    """
    This form is used to make it easy to enter results for a match.
    """

    def __init__(self, *args, **kwargs):
        super(MatchResultForm, self).__init__(*args, **kwargs)
        home_team = self.instance.home_team
        away_team = self.instance.away_team
        if self.instance.is_bye:
            self.fields.pop("home_team_score")
            self.fields.pop("away_team_score")
            self.fields.pop("is_forfeit")
            self.fields.pop("forfeit_winner")
            self.fields["bye_processed"].label = str(home_team or away_team)
        else:
            if home_team and away_team:
                self.fields["home_team_score"].label = str(home_team)
                self.fields["home_team_score"].widget.attrs["placeholder"] = home_team
                self.fields["away_team_score"].label = str(away_team)
                self.fields["away_team_score"].widget.attrs["placeholder"] = away_team
                self.fields["forfeit_winner"] = ModelChoiceField(
                    queryset=Team.objects.filter(id__in=[home_team.pk, away_team.pk]),
                    label_from_instance=lambda team: team.title,
                    empty_label=_("Double forfeit"),
                    help_text=_("Select the winning team."),
                    required=False,
                )
                # We have to manually set the class to form-control as this
                # isn't a model field.
                self.fields["forfeit_winner"].widget.attrs.setdefault(
                    "class", "form-control"
                )
            else:
                self.fields.pop("is_forfeit")
                self.fields.pop("forfeit_winner")
            self.fields.pop("bye_processed")

    def clean_forfeit_winner(self):
        if not self.cleaned_data.get("is_forfeit"):
            return None
        return self.cleaned_data.get("forfeit_winner")

    def clean(self):
        home_team_score = self.cleaned_data.get("home_team_score")
        away_team_score = self.cleaned_data.get("away_team_score")
        if home_team_score is not None and away_team_score is None:
            self.add_error("away_team_score", _("Both scores are required."))
        if away_team_score is not None and home_team_score is None:
            self.add_error("home_team_score", _("Both scores are required."))
        return self.cleaned_data

    def save(self, *args, **kwargs):
        logger.debug(
            'MatchResultForm.save: updated fields "%s"',
            '", "'.join(SCORE_FIELDS.intersection(self.changed_data)),
        )

        if SCORE_FIELDS.intersection(self.changed_data):
            for rec, res in score_updated.send_robust(sender=self, match=self.instance):
                receiver = "%s.%s" % (rec.__module__, rec.__name__)
                try:
                    if isinstance(res, Exception):
                        raise res
                    logger.debug("%s: %r", receiver, res)
                except:  # noqa
                    logger.exception('Receiver "%s" did not complete.', receiver)

        return super(MatchResultForm, self).save(*args, **kwargs)

    class Meta:
        model = Match
        fields = (
            "home_team_score",
            "away_team_score",
            "bye_processed",
            "is_forfeit",
            "forfeit_winner",
        )


MatchResultFormSet = modelformset_factory(Match, extra=0, form=MatchResultForm)


class MatchWashoutForm(BootstrapFormControlMixin, ModelForm):
    class Meta:
        model = Match
        fields = ("is_washout",)


MatchWashoutFormSet = modelformset_factory(Match, extra=0, form=MatchWashoutForm)


class MatchScheduleForm(BaseMatchFormMixin, ModelForm):
    def __init__(self, ignore_clashes=False, *args, **kwargs):
        super(MatchScheduleForm, self).__init__(*args, **kwargs)
        self.ignore_clashes = ignore_clashes

        tzinfo = timezone.get_current_timezone()
        __, tzname = timezone_choice(tzinfo.zone)

        def label_from_instance(obj):
            try:
                label = "-" * 3 + " " + obj.ground.title
            except ObjectDoesNotExist:
                label = obj.venue.title
            if settings.USE_TZ:
                __, tz = timezone_choice(obj.timezone)
                if tz != tzname:
                    label += " ({tz})".format(tz=tz)
            return label

        venues = collections.OrderedDict()
        for venue in self.instance.stage.division.season.venues.all():
            venues.setdefault(venue, [])

        for ground in Ground.objects.filter(venue__in=venues):
            venues[ground.venue].append(ground)

        places = FauxQueryset(Place)
        for venue, grounds in venues.items():
            places.append(venue)
            for ground in grounds:
                places.append(ground)

        self.fields["play_at"].queryset = places
        self.fields["play_at"].label_from_instance = label_from_instance
        if not settings.USE_TZ:
            self.fields["play_at"].help_text = (
                _(
                    "If not set, the timezone "
                    "for this match will be "
                    "assumed to be "
                    "<em>%s</em>."
                )
                % tzname
            )

    class Meta:
        model = Match
        fields = (
            "time",
            "play_at",
        )

    def clean_time(self):
        time = self.cleaned_data.get("time")
        if not self.ignore_clashes:
            for field in ("home", "away"):
                team = getattr(self.instance, field + "_team")
                after = getattr(team, "timeslots_after", None)
                before = getattr(team, "timeslots_before", None)
                # Check for team time preferences, does not rely on other
                # forms so can be checked at the form level rather than
                # formset level.
                if time and after and time < after:
                    raise forms.ValidationError(
                        _("%(team)s must play after %(time)s."),
                        code="invalid",
                        params={
                            "team": team.title,
                            "time": after.strftime("%H:%M"),
                        },
                    )
                elif time and before and time > before:
                    raise forms.ValidationError(
                        _("%(team)s must play before %(time)s."),
                        code="invalid",
                        params={
                            "team": team.title,
                            "time": before.strftime("%H:%M"),
                        },
                    )
        return time


class MatchScheduleManagementForm(ManagementForm):
    """
    Custom ``ManagementForm`` which adds a field that can control the clean
    method of the bound formset.

    When we want to skip the clash validation checking we can allow this - for
    example when it will be impossible to save the form otherwise.
    """

    ignore_clashes = forms.BooleanField(initial=0)


BaseMatchScheduleFormSet = modelformset_factory(Match, extra=0, form=MatchScheduleForm)


class MatchScheduleFormSet(BaseMatchScheduleFormSet):
    def _management_form(self):
        """
        Taken directly from ``django.forms.formsets.BaseFormSet`` and adjusted
        to make use of our custom ``ManagementForm`` subclass.

        TODO: provide a patch to upstream Django which allows a formset to
              configure, by way of class attribute, the ManagementForm class
              to instantiate.
        """
        if self.is_bound:
            form = MatchScheduleManagementForm(
                data=self.data, auto_id=self.auto_id, prefix=self.prefix
            )
            if not form.is_valid():
                raise forms.ValidationError(
                    "MatchScheduleManagementForm data "
                    "is missing or has been tampered "
                    "with"
                )
        else:
            form = MatchScheduleManagementForm(
                auto_id=self.auto_id,
                prefix=self.prefix,
                initial={
                    TOTAL_FORM_COUNT: self.total_form_count(),
                    INITIAL_FORM_COUNT: self.initial_form_count(),
                    MAX_NUM_FORM_COUNT: self.max_num,
                },
            )
        return form

    management_form = property(_management_form)

    def _construct_form(self, i, **kwargs):
        # obtain the ManagementForm.cleaned_data or fake empty
        mfcd = getattr(self.management_form, "cleaned_data", {})
        # determine form properties
        ignore_clashes = bool(mfcd.get("ignore_clashes", False))
        # construct the form with our additional keyword arguments
        return super(MatchScheduleFormSet, self)._construct_form(
            i, ignore_clashes=ignore_clashes, **kwargs
        )

    def clean(self):
        if self.management_form.cleaned_data.get("ignore_clashes"):
            return

        if any(self.errors):
            return

        teams = {}
        scheduled = {}

        # Build up a table of all possible conflicts
        dates = self.queryset.dates("date", "day")
        for m in self.queryset.model._default_manager.exclude(
            pk__in=self.queryset
        ).exclude(play_at=None, time=None):
            if m.date in dates:
                teams.setdefault(m.home_team, []).append(m.time)
                teams.setdefault(m.away_team, []).append(m.time)
                scheduled.setdefault((m.play_at, m.time), []).append(m)

        for i in range(0, self.total_form_count()):
            match = self.forms[i].cleaned_data.get("id")
            play_at = self.forms[i].cleaned_data.get("play_at")
            time = self.forms[i].cleaned_data.get("time")

            if play_at and time:
                key = (play_at, time)
                if scheduled.get(key):
                    err = _(
                        "Another match is already scheduled for " "this time & place."
                    )
                    self.forms[i].add_error("play_at", err)
                scheduled.setdefault(key, []).append(match)

            if match.home_team and time:
                for t in match.home_team.team_clashes.all():
                    if time in teams.get(t, []):
                        self.forms[i].add_error("time", t)
                teams.setdefault(match.home_team, []).append(time)

            if match.away_team and time:
                for t in match.away_team.team_clashes.all():
                    if time in teams.get(t, []):
                        self.forms[i].add_error("time", t)
                teams.setdefault(match.away_team, []).append(time)


class RescheduleDateForm(forms.Form):
    def __init__(self, matches, date, *args, **kwargs):
        self.original = date
        self.matches = matches.filter(date=date)
        super(RescheduleDateForm, self).__init__(*args, **kwargs)
        self.fields["date"] = SelectDateField(initial=date)

    def clean(self):
        errors = {}

        try:
            super(RescheduleDateForm, self).clean()
        except ValidationError as e:
            e.update_error_dict(errors)

        value = self.cleaned_data.get("date")
        representative_match = self.matches.earliest("date")

        if value < representative_match.stage.division.season.start_date:
            errors.setdefault("date", []).append(
                "This date is before the start of the season"
            )

        if errors:
            raise forms.ValidationError(errors)

    def save(self, *args, **kwargs):
        date = self.cleaned_data.get("date", self.original)
        if self.original is not date:
            return (date, self.matches.values_list("pk", flat=True))


RescheduleDateFormSetBase = formset_factory(RescheduleDateForm)


class RescheduleDateFormSet(ConstructFormMixin, RescheduleDateFormSetBase):
    def __init__(self, matches, dates, *args, **kwargs):
        self.matches = matches
        self.dates = dates
        super(RescheduleDateFormSet, self).__init__(*args, **kwargs)

    def total_form_count(self):
        return len(self.dates)

    def get_defaults(self):
        return {"matches": self.matches}

    def _construct_form(self, i, **kwargs):
        return super(RescheduleDateFormSet, self)._construct_form(
            i, date=self.dates[i], **kwargs
        )

    def save(self, *args, **kwargs):
        update = {}
        for form in self.forms:
            try:
                date, pks = form.save()
                update.setdefault(date, set()).update(pks)
            except TypeError:
                pass
        count = 0
        for date, pks in update.items():
            for match in self.matches.filter(match_unplayed, pk__in=pks):
                match.date = date
                match.clean()
                match.save()
                count += 1
        return count


class ProgressMatchesForm(BaseMatchFormMixin, ModelForm):
    def __init__(self, instance, *args, **kwargs):
        super(ProgressMatchesForm, self).__init__(instance=instance, *args, **kwargs)
        instance.evaluated = True
        home_team, away_team = instance.eval()

        # lazily move this into the init so we can remove RedefineModelForm
        self.fields["home_team"].empty_label = ""
        self.fields["away_team"].empty_label = ""

        # set initial form values when available, prevents need for client
        # side javascript hack used previously.
        self.initial["home_team"] = home_team
        self.initial["away_team"] = away_team

        if (
            isinstance(home_team, ByeTeam)
            or self.instance.home_team
            or home_team is None
        ):
            self.fields.pop("home_team", None)
        elif self.instance.home_team_eval_related:
            self.fields["home_team"].queryset = self.fields[
                "home_team"
            ].queryset.filter(
                pk__in=(
                    self.instance.home_team_eval_related.home_team_id,
                    self.instance.home_team_eval_related.away_team_id,
                )
            )

        if (
            isinstance(away_team, ByeTeam)
            or self.instance.away_team
            or away_team is None
        ):
            self.fields.pop("away_team", None)
        elif self.instance.away_team_eval_related:
            self.fields["away_team"].queryset = self.fields[
                "away_team"
            ].queryset.filter(
                pk__in=(
                    self.instance.away_team_eval_related.home_team_id,
                    self.instance.away_team_eval_related.away_team_id,
                )
            )

    def has_changed(self):
        # Because we're setting the initial directly in the __init__ above we
        # need to always mark the data as changed.
        return True

    def save(self, *args, **kwargs):
        self.instance.stage.matches_needing_printing.add(self.instance)
        return super(ProgressMatchesForm, self).save(*args, **kwargs)

    class Meta:
        model = Match
        fields = (
            "home_team",
            "away_team",
        )


ProgressMatchesFormSet = modelformset_factory(Match, form=ProgressMatchesForm, extra=0)


class ProgressTeamsForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(ProgressTeamsForm, self).__init__(*args, **kwargs)
        teams = self.instance.choices.order_by("title")
        self.fields["team"] = ModelChoiceField(
            queryset=teams, label_from_instance="title"
        )

    def save(self, *args, **kwargs):
        team = self.cleaned_data.get("team")
        self.instance.home_games.update(home_team=team)
        self.instance.away_games.update(away_team=team)
        return self.instance

    class Meta:
        model = UndecidedTeam
        fields = ("id",)


BaseProgressTeamsFormSet = modelformset_factory(
    UndecidedTeam, form=ProgressTeamsForm, extra=0
)


class ProgressTeamsFormSet(BaseProgressTeamsFormSet):
    def __init__(self, stage, *args, **kwargs):
        self.stage = stage
        super(ProgressTeamsFormSet, self).__init__(*args, **kwargs)

    def get_queryset(self):
        qs = super(ProgressTeamsFormSet, self).get_queryset()
        return qs.filter(stage=self.stage)

    def clean(self):
        teams = {}

        for i, form in enumerate(self.forms):
            if not hasattr(form, "cleaned_data"):
                continue
            teams.setdefault(form.cleaned_data.get("team"), []).append(i)

        for team, rows in teams.items():
            if len(rows) > 1:
                for i in rows:
                    self.forms[i].add_error(
                        "team", _("You can only progress a team into one position.")
                    )

    def save(self, *args, **kwargs):
        res = super(ProgressTeamsFormSet, self).save(*args, **kwargs)
        self.stage.matches_needing_printing = self.stage.matches.exclude(
            legitimate_bye_match
        )
        return res


class DrawGenerationForm(BootstrapFormControlMixin, forms.Form):

    start_date = SelectDateField()
    format = ModelChoiceField(queryset=DrawFormat.objects.all())
    rounds = forms.IntegerField(required=False, min_value=1)
    offset = forms.IntegerField(required=False)

    def __init__(self, initial, *args, **kwargs):
        super(DrawGenerationForm, self).__init__(*args, **kwargs)
        self.instance = initial

        # ensure we have an even number for filtering the `DrawFormat` table

        if isinstance(self.instance, Stage):
            teams = first(
                (self.instance.teams.count(), self.instance.undecided_teams.count()),
                default=0,
            )

        elif isinstance(self.instance, StageGroup):
            teams = first(
                (self.instance.undecided_teams.count(), self.instance.teams.count()),
                default=0,
            )

        else:
            teams = 0

        if teams % 2:
            teams += 1

        # produce a list of appropriate `DrawFormat` options
        suitable_draw_formats = DrawFormat.objects.filter(
            Q(teams__in=(teams, teams - 1)) if teams else Q()
        )
        self.fields["format"].queryset = suitable_draw_formats
        self.fields["format"].help_text = _(
            "Choose the competition format for this %s"
        ) % (self.instance._meta.verbose_name.lower(),)

        # determine the best logical start date
        start_date = self.instance.division.season.start_date or datetime.date.today()
        self.fields["start_date"].initial = start_date

    @property
    def generator(self):
        format = self.cleaned_data.get("format")
        start_date = self.cleaned_data.get("start_date")
        return format.generator(self.instance, start_date)

    def clean_start_date(self):
        start_date = self.cleaned_data.get("start_date")
        if not start_date:
            raise forms.ValidationError("foo")
        return start_date

    def clean_rounds(self):
        format = self.cleaned_data.get("format")
        rounds = self.cleaned_data.get("rounds")
        if not rounds and format:
            rounds = len(self.generator.rounds)
        return rounds

    def clean(self):
        data = super(DrawGenerationForm, self).clean()
        try:
            matches = self.get_matches()
        except AttributeError:
            logger.exception("Can't build draw without a format selection.")
        else:
            data.setdefault("matches", matches)
        return data

    def get_matches(self):
        n = self.cleaned_data.get("rounds")
        offset = self.cleaned_data.get("offset") or 0
        return self.generator.generate(n, offset)


DrawGenerationFormSetBase = formset_factory(
    form=DrawGenerationForm, extra=0, can_delete=True
)


class DrawGenerationFormSet(BootstrapFormControlMixin, DrawGenerationFormSetBase):
    def add_fields(self, form, index):
        """A hook for adding extra fields on to each form instance."""
        super(DrawGenerationFormSet, self).add_fields(form, index)
        if self.can_delete:
            form.fields[DELETION_FIELD_NAME] = BooleanChoiceField(
                label=_("Skip"), required=False
            )


class ImportCsvForm(forms.Form):
    csv = forms.FileField(
        label=_("File"),
        help_text=_(
            "Any columns in your data file that " "are unknown will be discarded."
        ),
    )

    def clean_csv(self):
        """
        The error handling could be better if we make use of `libmagic` but
        hopefully this will prevent the vast majority of failures.
        """
        csv = self.cleaned_data.get("csv")
        main, subtype = csv.content_type.split("/", 1)
        if main != "text":
            raise forms.ValidationError(_("You must submit a comma separated file."))
        return csv


class BaseClubAssociationFormSet(UserMixin, ConstructFormMixin, BaseInlineFormSet):
    def get_defaults(self):
        return {"user": self.user, "club": self.instance}


class BaseTeamAssociationFormSet(UserMixin, ConstructFormMixin, BaseInlineFormSet):
    def get_defaults(self):
        return {"user": self.user, "team": self.instance}


class BaseSeasonAssociationFormSet(UserMixin, ConstructFormMixin, BaseModelFormSet):
    def __init__(self, club, season, *args, **kwargs):
        self.club = club
        self.season = season
        super(BaseSeasonAssociationFormSet, self).__init__(*args, **kwargs)

    def get_defaults(self):
        return {"user": self.user, "club": self.club, "season": self.season}


class ClubAssociationForm(UserMixin, ModelForm):
    def __init__(self, club, *args, **kwargs):
        super(ClubAssociationForm, self).__init__(*args, **kwargs)
        self.fields["person"].queryset = club.members.all()
        if self.instance.person_id is None:
            self.fields["person"].required = False

    class Meta:
        model = ClubAssociation
        fields = (
            "person",
            "roles",
        )


class TeamAssociationForm(UserMixin, ModelForm):
    def __init__(self, team, *args, **kwargs):
        super(TeamAssociationForm, self).__init__(*args, **kwargs)
        self.fields["person"].queryset = team.club.members.all()
        self.fields[
            "roles"
        ].queryset = team.division.season.competition.team_roles.all()

    class Meta:
        model = TeamAssociation
        fields = (
            "number",
            "person",
            "is_player",
            "roles",
        )


class SeasonAssociationForm(UserMixin, ModelForm):
    def __init__(self, club, season, *args, **kwargs):
        super(SeasonAssociationForm, self).__init__(*args, **kwargs)
        self.fields["person"].queryset = club.members.all()
        self.fields["roles"].queryset = season.competition.club_roles.all()

        self.instance.club = club
        self.instance.season = season

    class Meta:
        model = SeasonAssociation
        fields = ("person", "roles")


ClubAssociationFormSet = inlineformset_factory(
    Club,
    ClubAssociation,
    form=ClubAssociationForm,
    formset=BaseClubAssociationFormSet,
    extra=0,
)

TeamAssociationFormSet = inlineformset_factory(
    Team,
    TeamAssociation,
    form=TeamAssociationForm,
    formset=BaseTeamAssociationFormSet,
    extra=0,
)

SeasonAssociationFormSet = modelformset_factory(
    SeasonAssociation,
    form=SeasonAssociationForm,
    formset=BaseSeasonAssociationFormSet,
    extra=0,
    can_delete=True,
)


class BaseSeasonExclusionFormSet(UserMixin, BaseInlineFormSet):
    pass


SeasonExclusionFormSet = inlineformset_factory(
    Season,
    SeasonExclusionDate,
    fields=("date",),
    formset=BaseSeasonExclusionFormSet,
    extra=0,
)


class BaseDivisionExclusionFormSet(UserMixin, BaseInlineFormSet):
    pass


DivisionExclusionFormSet = inlineformset_factory(
    Division,
    DivisionExclusionDate,
    fields=("date",),
    formset=BaseDivisionExclusionFormSet,
    extra=0,
)


class ClubRoleForm(BootstrapFormControlMixin, ModelForm):
    class Meta:
        model = ClubRole
        fields = ("name",)


class TeamRoleForm(BootstrapFormControlMixin, ModelForm):
    class Meta:
        model = TeamRole
        fields = ("name",)


class SimpleScoreMatchStatisticForm(ModelForm):
    class Meta:
        model = SimpleScoreMatchStatistic
        fields = ("played", "number", "points", "mvp")
        widgets = {
            "played": MatchPlayedWidget,
        }

    def has_changed(self):
        """
        Ensure that the entry is always updated, otherwise the default of
        played won't result in a cap being awarded *unless* one of the other
        fields is populated.
        """
        return True

    def clean_points(self):
        points = self.cleaned_data["points"]
        if points is None:
            return 0
        return points


BaseMatchStatisticFormset = modelformset_factory(
    SimpleScoreMatchStatistic,
    form=SimpleScoreMatchStatisticForm,
    extra=0,
)


class MatchStatisticFormset(BaseMatchStatisticFormset):
    def __init__(self, score, *args, **kwargs):
        super(MatchStatisticFormset, self).__init__(*args, **kwargs)
        self.score = score

    def _construct_form(self, i, **kwargs):
        if i < self.initial_form_count() and not kwargs.get("instance"):
            kwargs["instance"] = self.get_queryset()[i]
        return super(BaseModelFormSet, self)._construct_form(i, **kwargs)

    def clean(self):
        if any(self.errors):
            raise forms.ValidationError(
                _("There are errors you must fix " "before we can verify the scores.")
            )

        if self.score is None:
            raise forms.ValidationError(
                _(
                    "You must set the simple match score prior to entering "
                    "detailed match statistics."
                )
            )

        score = sum([f.cleaned_data["points"] for f in self.forms])
        if score != self.score:
            raise forms.ValidationError(
                _(
                    "Total number of points (%(points)d) does not equal total "
                    "number of scores (%(scores)d) for this team."
                )
                % {"points": score, "scores": self.score}
            )

        players = len([f for f in self.forms if f.cleaned_data["played"]])
        maximum = 14  # FIXME this maximum should not be hard-coded
        if players > maximum:
            message = (
                ngettext(
                    "A maximum of %(max)d players may participate in a match, "
                    "there is %(count)d selected.",
                    "A maximum of %(max)d players may participate in a match, "
                    "there are %(count)d selected.",
                    players,
                )
                % {"max": maximum, "count": players}
            )
            raise forms.ValidationError(message)

    def save(self, *args, **kwargs):
        # Should look into the correct solution. I think it should be to overload save_new
        # on the base Form to attach it to the team... that seems to be the only reason we
        # need to pre-populate with a "faux queryset" in the FormSet?
        stats = []
        for form in self.forms:
            stats.append(form.save())
        return stats


SeasonMatchTimeFormSet = inlineformset_factory(
    Season,
    SeasonMatchTime,
    fk_name="season",
    fields=(
        "start_date",
        "end_date",
        "start",
        "interval",
        "count",
    ),
    extra=0,
    can_delete=True,
)


class TournamentScheduleForm(BootstrapFormControlMixin, forms.Form):

    team_hook = forms.CharField(
        widget=forms.Textarea,
        label="Teams",
        help_text="Enter team names, one per line, in seeding order.",
    )
    days_available = forms.IntegerField(
        min_value=1,
        label="How many days of competition are there?",
        help_text="Including the final series.",
    )
    max_per_day = forms.IntegerField(
        initial=3,
        min_value=0,
        label="What is the daily maximum number of games a team may play?",
    )
    min_per_day = forms.IntegerField(
        initial=1,
        min_value=0,
        label="What is the daily minimum number of games a team must play?",
        help_text="Only applies to preliminary stages.",
    )

    def get_teams(self):
        return [line for line in self.cleaned_data["team_hook"].splitlines() if line]

    def get_context(self):
        assert self.is_valid()

        seeded_teams = self.get_teams()
        days_available = self.cleaned_data["days_available"]
        max_per_day = self.cleaned_data["max_per_day"]
        min_per_day = self.cleaned_data["min_per_day"]

        context = seeded_tournament(
            seeded_teams, days_available, max_per_day, min_per_day
        )
        return context


class DivisionTournamentScheduleForm(TournamentScheduleForm):

    team_hook = ModelChoiceField(
        queryset=(
            Division.objects.filter(teams__isnull=False)
            .order_by("season__competition", "season")
            .distinct()
            .select_related()
        ),
        label_from_instance=lambda o: "{} - {} - {}".format(
            o.season.competition.title,
            o.season.title,
            o.title,
        ),
        label="Division",
    )

    def get_absolute_url(self):
        division = self.cleaned_data["team_hook"]
        if division.matches.count():
            return reverse_lazy(
                "competition:division",
                kwargs=dict(
                    competition=division.season.competition.slug,
                    season=division.season.slug,
                    division=division.slug,
                ),
            )

    def get_teams(self):
        return self.cleaned_data["team_hook"].teams.order_by("order")



from django.conf import settings

__all__ = (
    'APP_ROUTING',
    'CURRENCY_ABBREVIATION',
    'CURRENCY_SYMBOL',
    'PAGINATE_BY',
    'SITEMAP_CACHE_DURATION',
    'SITEMAP_EDIT_PARENT',
    'SITEMAP_HTTPS_OPTION',
    'SITEMAP_ROOT',
    'STORAGE_FOLDER',
    'STORAGE_URL',
)


def A(n, d):
    return getattr(settings, 'AUTHENTICATION_' + n, d)


def S(n, d=None):
    return getattr(settings, 'TOUCHTECHNOLOGY_' + n, d)


APP_ROUTING = S('APP_ROUTING', ())
CURRENCY_ABBREVIATION = S('CURRENCY_ABBREVIATION', 'AUD')
CURRENCY_SYMBOL = S('CURRENCY_SYMBOL', '$')
PAGINATE_BY = S('PAGINATE_BY', 5)
PROFILE_FORM_CLASS = S('PROFILE_FORM_CLASS', 'touchtechnology.common.forms_lazy.ProfileForm')
SITEMAP_CACHE_DURATION = S('SITEMAP_DURATION')
SITEMAP_EDIT_PARENT = S('SITEMAP_EDIT_PARENT', False)
SITEMAP_HTTPS_OPTION = S('HTTPS_OPTION', False)
SITEMAP_ROOT = S('SITEMAP_ROOT')
STORAGE_FOLDER = S('STORAGE_FOLDER')
STORAGE_URL = S('STORAGE_URL')
